"""Implements algorithms for clustering scans by
spectral similarity.
"""
import operator
import functools
import bisect
import json

from typing import Dict, List, Optional, Tuple, Union
from collections import deque

import numpy as np
from scipy.special import comb

from ms_deisotope.task.log_utils import LogUtilsMixin

from ms_deisotope.data_source.scan import ScanBase
from ms_deisotope.data_source.dispatch import (
    SubsequenceMappingProxy as _SubsequenceMappingProxy,
    DynamicallyLoadingProxyResolver as _DynamicallyLoadingResolver)

from .similarity_methods import peak_set_similarity

try:
    basestring
except NameError:
    from six import string_types as basestring

peak_set_getter = operator.attrgetter("peak_set")
deconvoluted_peak_set_getter = operator.attrgetter("deconvoluted_peak_set")


def _ppm_error(x, y):
    return (x - y) / y


@functools.total_ordering
class SpectrumCluster(object):
    """A collection of similar spectra.

    A :class:`SpectrumCluster` is compare-able by :attr:`neutral_mass`
    and acts as a :class:`collections.Sequence` over its members stored
    in :attr:`scans`.

    Attributes
    ----------
    scans: :class:`list`
        A list of spectra which are similar to each other.
    neutral_mass: :class:`float`
        The neutral mass of the representative spectrum's precursor.
    """
    scans: List[ScanBase]
    neutral_mass: float
    _average_similarity: Optional[float]
    annotations: Dict

    def __init__(self, scans=None, neutral_mass=None, average_similarity=None, annotations=None):
        if scans is None:
            scans = []
        if annotations is None:
            annotations = dict()
        self.scans = scans
        self.neutral_mass = None
        if neutral_mass:
            self.neutral_mass = neutral_mass
        elif self.scans:
            self.neutral_mass = self.scans[0].precursor_information.neutral_mass
        else:
            self.neutral_mass = 0.0
        if average_similarity is None and len(self) == 1:
            average_similarity = 1.0
        self._average_similarity = average_similarity
        self.annotations = annotations

    def __lt__(self, other: 'SpectrumCluster'):
        return self.neutral_mass < other.neutral_mass

    def __gt__(self, other: 'SpectrumCluster'):
        return self.neutral_mass > other.neutral_mass

    def __eq__(self, other: 'SpectrumCluster'):
        return (abs(self.neutral_mass - other.neutral_mass) / other.neutral_mass) < 1e-6

    def __repr__(self):
        return "SpectrumCluster(%f, %d)" % (self.neutral_mass, len(self))

    def __iter__(self):
        return iter(self.scans)

    def __len__(self):
        return len(self.scans)

    def __getitem__(self, i):
        return self.scans[i]

    def _invalidate(self):
        self._average_similarity = None

    def append(self, item: ScanBase, incremental_similarity=False):
        """Add a new spectrum to the cluster.

        Parameters
        ----------
        item: :class:`~.ScanBase`
        """
        if not self.neutral_mass:
            self.neutral_mass = item.precursor_information.neutral_mass
        if incremental_similarity:
            self._incremental_similarity(item)
        else:
            self._invalidate()
        self.scans.append(item)

    def _calculate_similarity_with(self, scan: ScanBase, from_ix=0, to_ix=None, *args, **kwargs) -> List[float]:
        if from_ix is None:
            from_ix = 0
        if to_ix is None:
            to_ix = len(self)
        acc = []
        for member in self[from_ix:to_ix]:
            acc.append(peak_set_similarity(scan, member, *args, **kwargs))
        return acc

    def _incremental_similarity(self, scan: ScanBase, *args, **kwargs):
        new_sims = self._calculate_similarity_with(scan, *args, **kwargs)
        aggregate_size = comb(len(self), 2)
        n = (aggregate_size + len(new_sims))
        if n == 0:
            n = 1
        self._average_similarity = (aggregate_size * self.average_similarity() + sum(new_sims)
                                   ) / n

    def _full_similarity(self, *args, **kwargs):
        similarity_method = kwargs.pop(
            "similarity_method", peak_set_similarity)
        ratings = []
        n = len(self)
        for i in range(n):
            scan_i = self[i]
            for j in range(i + 1, n):
                scan_j = self[j]
                ratings.append(similarity_method(
                    scan_i, scan_j, *args, **kwargs))
        self._average_similarity = sum(ratings) / len(ratings)

    def average_similarity(self, *args, **kwargs) -> float:
        """Calculate the within-cluster similarity among all cluster members
        and returns the average.

        All arguments are forwarded to :func:`~.peak_set_similarity`.

        If the cluster is a singleton or smaller, by definition its similarity
        is ``1.0``.

        Parameters
        ----------
        similarity_method: Callable, optional
            The peak set similarity method to use. Defaults to :func:`~.peak_set_similarity`
            All other arguments are forwarded to this function.

        Returns
        -------
        :class:`float`
        """
        n = len(self)
        if n <= 1:
            return 1.0
        if self._average_similarity is not None:
            return self._average_similarity
        self._full_similarity(*args, **kwargs)
        return self._average_similarity

    def similarity_matrix(self, *args, **kwargs):
        """Compute a symmetric similarity matrix comparing each
        spectrum within the cluster to each other spectrum.

        Parameters
        ----------
        similarity_method: Callable, optional
            The peak set similarity method to use. Defaults to :func:`~.peak_set_similarity`
            All other arguments are forwarded to this function.

        Returns
        -------
        :class:`np.ndarray`: m
            The value at position ``m[i,j]`` is the simiarlity of ``self[i]`` with ``self[j]``
        """
        similarity_method = kwargs.pop("similarity_method", peak_set_similarity)
        n = len(self)
        mat = np.identity(n)
        for i in range(n):
            scan_i = self[i]
            for j in range(i + 1, n):
                scan_j = self[j]
                mat[i, j] = mat[j, i] = (similarity_method(scan_i, scan_j, *args, **kwargs))
        return mat

    def to_dict(self):
        """Convert the cluster to a JSON-safe :class:`dict`

        Returns
        -------
        :class:`dict`
        """
        d = {}
        d['neutral_mass'] = self.neutral_mass
        d['size'] = len(self)
        d['average_similarity'] = self.average_similarity()
        scans = []
        for scan in self:
            scan_source = scan.source
            if scan_source is not None:
                source_name = scan_source.source_file
                if not isinstance(source_name, basestring):
                    if hasattr(source_name, 'name'):
                        source_name = source_name.name
            else:
                source_name = ":detatched:"
            scans.append({
                "id": scan.id,
                "source": source_name,
                "neutral_mass": scan.precursor_information.neutral_mass,
            })
        d['scans'] = scans
        return d

    def split_on_charge(self) -> Dict[int, 'SpectrumCluster']:
        scans = sorted(self, key=lambda x: x.tic(), reverse=True)
        index = {}
        for scan in scans:
            key = scan.precursor_information.charge
            try:
                index[key].append(scan)
            except KeyError:
                index[key] = SpectrumCluster([scan], scan.precursor_information.neutral_mass)
        return index


class SpectrumClusterCollection(object):
    """A sorted :class:`~.Sequence` of :class:`SpectrumCluster` instances
    that supports searching by precursor mass.
    """
    clusters: List[SpectrumCluster]

    def __init__(self, clusters=None):
        if clusters is None:
            clusters = []
        self.clusters = list(clusters)

    def add(self, cluster: SpectrumCluster):
        """Add a new :class:`SpectrumCluster` to the collection,
        preserving sorted order.

        Parameters
        ----------
        cluster: :class:`SpectrumCluster`
            The cluster to add
        """
        bisect.insort(self.clusters, cluster)

    def __getitem__(self, i):
        return self.clusters[i]

    def __setitem__(self, i, value):
        self.clusters[i] = value

    def __len__(self):
        return len(self.clusters)

    def __iter__(self):
        return iter(self.clusters)

    def __repr__(self):
        template = "{self.__class__.__name__}({size})"
        size = len(self)
        return template.format(self=self, size=size)

    def _binary_search(self, mass: float, error_tolerance: float=1e-5) -> Tuple[int, int, int]:
        array = self.clusters
        n = len(array)
        lo = 0
        hi = n

        while hi != lo:
            mid = (hi + lo) // 2
            y = array[mid].neutral_mass
            err = (y - mass) / mass
            if hi - lo == 1:
                best_index = mid
                best_error = abs(err)
                i = mid - 1
                while i >= 0:
                    x = array[i]
                    err = abs((x.neutral_mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1
                lo_index = i + 1
                i = mid + 1
                while i < n:
                    x = array[i]
                    err = abs((x.neutral_mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                hi_index = i
                return best_index, lo_index, hi_index
            elif err > 0:
                hi = mid
            else:
                lo = mid
        return 0, 0, 0

    def find(self, mass: float, error_tolerance: float=1e-5) -> Optional[SpectrumCluster]:
        """Finds the cluster whose precursor mass is closest to
        ``mass`` within ``error_tolerance`` ppm error.

        Parameters
        ----------
        mass: :class:`float`
            The mass to search for.
        error_tolerance: :class:`float`
            The PPM error tolerance to use. Defaults to 1e-5.

        Returns
        -------
        :class:`SpectrumCluster` or :const:`None`
        """
        target_ix, _, _ = self._binary_search(mass, error_tolerance)
        target = self[target_ix]
        if abs(target.neutral_mass - mass) / mass > error_tolerance:
            return None
        return target

    def find_all(self, mass: float, error_tolerance: float=1e-5) -> List[SpectrumCluster]:
        """Finds all clusters whose precursor mass is within ``error_tolerance``
        ppm of ``mass``.

        Parameters
        ----------
        mass: :class:`float`
            The mass to search for.
        error_tolerance: :class:`float`
            The PPM error tolerance to use. Defaults to 1e-5.

        Returns
        -------
        :class:`list`
        """
        _, lo_ix, hi_ix = self._binary_search(mass, error_tolerance)
        result = [
            target for target in self[lo_ix:hi_ix]
            if abs(target.neutral_mass - mass) / mass <= error_tolerance
        ]
        return result


class _PeakGetterStrategyBase(object):
    def __call__(self, scan):
        return self.peaks(scan)


class _FittedPeakAccessorStrategy(_PeakGetterStrategyBase):
    def peaks(self, scan):
        return scan.peak_set

    def tic(self, scan):
        return scan.tic.centroided()


class _DeconvolutedPeakAccessorStrategy(_PeakGetterStrategyBase):
    def peaks(self, scan):
        return scan.deconvoluted_peak_set

    def tic(self, scan):
        return scan.tic.deconvoluted()


class _DynamicPeakAccessorStrategy(_PeakGetterStrategyBase):

    def _ensure_peak_set(self, scan):
        if scan.peak_set is None:
            scan.pick_peaks()
        return scan.peak_set

    def peaks(self, scan):
        if scan.deconvoluted_peak_set is not None:
            return scan.deconvoluted_peak_set
        else:
            return self._ensure_peak_set(scan)

    def tic(self, scan):
        if scan.deconvoluted_peak_set is not None:
            return scan.tic.deconvoluted()
        else:
            self._ensure_peak_set(scan)
            return scan.tic.centroided()


def group_by_precursor_mass(scans, window_size=1.5e-5):
    scans = sorted(
        scans, key=lambda x: x.precursor_information.extracted_neutral_mass,
        reverse=True)
    groups = []
    if len(scans) == 0:
        return groups
    current_group = [scans[0]]
    last_scan = scans[0]
    for scan in scans[1:]:
        delta = abs(
            (scan.precursor_information.extracted_neutral_mass -
             last_scan.precursor_information.extracted_neutral_mass
             ) / last_scan.precursor_information.extracted_neutral_mass)
        if delta > window_size:
            groups.append(current_group)
            current_group = [scan]
        else:
            current_group.append(scan)
        last_scan = scan
    groups.append(current_group)
    return groups


class ScanClusterBuilder(LogUtilsMixin):
    """Clusters spectra based upon peak pattern similarity

    Attributes
    ----------
    clusters : :class:`SpectrumClusterCollection`
        The clusters built so far
    minimum_similarity : float
        The minimum similarity score needed to consider two spectra
        similar enough to form a cluster
    peak_getter : :class:`Callable`
        A function to call on each spectrum to retrieve the peak list
        to cluster over
    precursor_error_tolerance : float
        The maximum precursor mass error (in PPM) to permit between
        two spectra to consider comparing them
    track_incremental_similarity : bool
        Whether to incrementally update a cluster's similarity when it
        grows.
    """

    @classmethod
    def _guess_peak_getter(cls, getter):
        if getter is None:
            return _DynamicPeakAccessorStrategy()
        if callable(getter):
            return getter
        if isinstance(getter, basestring):
            if getter == "d":
                return _DeconvolutedPeakAccessorStrategy()
            if getter in ('p', 'c'):
                return _FittedPeakAccessorStrategy()
            else:
                raise ValueError("Cannot infer peak getter strategy from %r" % (getter, ))

        raise ValueError(
            "Cannot infer peak set getter strategy from %r" % (getter, ))

    def __init__(self, clusters=None, precursor_error_tolerance=1e-5, minimum_similarity=0.1,
                 peak_getter=None, track_incremental_similarity=True):
        peak_getter = self._guess_peak_getter(peak_getter)
        if clusters is None:
            clusters = []
        self.clusters = SpectrumClusterCollection(clusters)
        self.precursor_error_tolerance = precursor_error_tolerance
        self.minimum_similarity = minimum_similarity
        self.peak_getter = peak_getter
        self.track_incremental_similarity = track_incremental_similarity

    def _binsearch_simple(self, x):
        n = len(self)
        lo = 0
        hi = n

        while hi != lo:
            mid = (hi + lo) // 2
            y = self[mid].neutral_mass
            err = y - x
            # Do refinement in `find_best_cluster_for_scan`
            if hi - lo == 1:
                return mid
            elif err > 0:
                hi = mid
            else:
                lo = mid
        return 0

    def peak_set_similarity(self, scan_i, scan_j):
        """Calculate the similarity between the peaks of
        ``scan_i`` and ``scan_j``, where the peaks are
        those points given by :attr:`peak_getter`

        Parameters
        ----------
        scan_i: :class`~.ScanBase`
        scan_j: :class`~.ScanBase`

        Returns
        -------
        :class:`float`
        """
        peak_set_a = self.peak_getter(scan_i)
        peak_set_b = self.peak_getter(scan_j)
        return peak_set_similarity(
            peak_set_a, peak_set_b)

    def find_best_cluster_for_scan(self, scan):
        """Locate the best cluster to add ``scan`` to according to
        precursor mass and peak set similarity.

        Parameters
        ----------
        scan: :class:`~.ScanBase`

        Returns
        -------
        :class:`SpectrumCluster`
        """
        best_cluster = None
        best_similarity = 0.0
        n = len(self.clusters)
        if n == 0:
            return best_cluster

        center_i = self._binsearch_simple(scan.precursor_information.neutral_mass)
        i = center_i

        while i >= 0:
            cluster = self.clusters[i]
            if abs(_ppm_error(scan.precursor_information.neutral_mass,
                              cluster.neutral_mass)) > self.precursor_error_tolerance:
                break
            similarity = self.peak_set_similarity(scan, cluster[0])
            i -= 1
            if similarity > best_similarity and similarity > self.minimum_similarity:
                best_similarity = similarity
                best_cluster = cluster

        i = center_i + 1
        while i < n:
            cluster = self.clusters[i]
            if abs(_ppm_error(scan.precursor_information.neutral_mass,
                              cluster.neutral_mass)) > self.precursor_error_tolerance:
                break
            similarity = self.peak_set_similarity(scan, cluster[0])
            i += 1
            if similarity > best_similarity and similarity > self.minimum_similarity:
                best_similarity = similarity
                best_cluster = cluster
        return best_cluster

    def add_scan(self, scan):
        """Add ``scan`` to the cluster collection, adding it to the best
        matching cluster, or starting a new cluster around it if no good
        match can be found.

        Parameters
        ----------
        scan: :class:`~.ScanBase`
        """
        best_cluster = self.find_best_cluster_for_scan(scan)
        if best_cluster:
            best_cluster.append(scan, incremental_similarity=self.track_incremental_similarity)
        else:
            self.clusters.add(SpectrumCluster([scan]))

    def __iter__(self):
        return iter(self.clusters)

    def __len__(self):
        return len(self.clusters)

    def __getitem__(self, i):
        return self.clusters[i]

    def _get_tic(self, scan):
        try:
            return self.peak_getter.tic(scan)
        except AttributeError:
            return sum(p.intensity for p in self.peak_getter(scan))

    @classmethod
    def cluster_scans(cls, scans, precursor_error_tolerance=1e-5, minimum_similarity=0.1,
                      peak_getter=None, sort=True, track_incremental_similarity=False):
        """Cluster scans by precursor mass and peak set similarity.

        Parameters
        ----------
        scans: :class:`Iterable`
            An iterable of :class:`Scan`-like objects
        precursor_error_tolerance: :class:`float`
            The PPM mass accuracy threshold for precursor mass differences to
            tolerate when deciding whether to compare spectra. Defaults to 1e-5.
        minimum_similarity: :class:`float`
            The minimum peak set similarity required to consider adding a spectrum
            to a cluster. Defaults to 0.1
        peak_getter: :class:`Callable`
            A callable object used to get peaks from elements of ``scans``.
        sort: :class:`bool`
            Whether or not to sort spectra by their total ion current before clustering.
        """
        self = cls([], precursor_error_tolerance, minimum_similarity,
                   peak_getter, track_incremental_similarity)
        if sort:
            scans = self._sort_by_tic(scans)
        if len(scans) > 100:
            self.log("Clustering %d Scans" % (len(scans), ))
        n = len(scans)
        report_interval = max(min(n // 10, 1000), 50)
        for i, scan in enumerate(scans):
            if i % report_interval == 0 and i:
                self.log("... Handled %d Scans (%0.2f%%)" % (i, i * 100.0 / n))
            self.add_scan(scan)
        return self.clusters

    def _sort_by_tic(self, scans):
        should_log = len(scans) > 100
        if should_log:
            self.log("Sorting Scans By TIC")
        augmented = []
        n = len(scans)
        for i, scan in enumerate(scans):
            if i % 1000 == 0 and i > 0:
                self.log("... Loaded TIC for %d Scans (%0.2f%%)" % (i, i * 100.0 / n))
            augmented.append((self._get_tic(scan), scan))
        augmented.sort(key=lambda x: x[0], reverse=True)
        scans = [a[1] for a in augmented]
        return scans

    @classmethod
    def iterative_clustering(cls, scans, precursor_error_tolerance=1e-5, similarity_thresholds=None,
                             peak_getter=None):
        """Cluster scans by precursor mass and peak set similarity, iteratively refining
        clusters with increasing similarity threshold requirements.

        Parameters
        ----------
        scans: :class:`Iterable`
            An iterable of :class:`Scan`-like objects
        precursor_error_tolerance: :class:`float`
            The PPM mass accuracy threshold for precursor mass differences to
            tolerate when deciding whether to compare spectra. Defaults to 1e-5.
        similarity_thresholds: :class:`Sequence` of :class:`float`
            A series of similarity thresholds to apply as spectra are added to clusters
            and as clusters are iteratively refined.
        peak_getter: :class:`Callable`
            A callable object used to get peaks from elements of ``scans``.
        """
        peak_getter = cls._guess_peak_getter(peak_getter)
        if similarity_thresholds is None:
            similarity_thresholds = [0.1, .4, 0.7]
        singletons = []
        to_bisect = [scans]
        logger = LogUtilsMixin()
        track_similarity = [False] * (len(similarity_thresholds) - 1)
        track_similarity.append(True)
        for similarity_threshold, track in zip(similarity_thresholds, track_similarity):
            logger.log("Clustering with Threshold %0.2f" % (similarity_threshold, ))
            next_to_bisect = []
            if len(to_bisect) > 1:
                logger.log("Refining %d Clusters" % (len(to_bisect)))
            elif to_bisect:
                logger.log("Clustering %d Scans" % (len(to_bisect[0])))
            else:
                logger.log("Nothing to cluster...")
                break
            n = len(to_bisect)
            report_interval = max(min(n // 10, 1000), 1)
            for i, group in enumerate(to_bisect):
                if i % report_interval == 0:
                    logger.log("... Handling Batch %d (%d Scans)" % (i, len(group)))
                clusters = cls.cluster_scans(
                    group, precursor_error_tolerance,
                    minimum_similarity=similarity_threshold,
                    peak_getter=peak_getter, sort=True, track_incremental_similarity=track)
                for cluster in clusters:
                    if len(cluster) == 1:
                        singletons.append(cluster)
                    else:
                        next_to_bisect.append(cluster)
            logger.log("%d Singletons and %d Groups" % (len(singletons), len(next_to_bisect)))
            to_bisect = next_to_bisect
        return SpectrumClusterCollection(sorted(list(singletons) + list(to_bisect)))


cluster_scans = ScanClusterBuilder.cluster_scans
iterative_clustering = ScanClusterBuilder.iterative_clustering


class ScanClusterWriterBase(object):
    """A base class for writing :class:`ScanCluster` objects to an
    I/O stream like a file.

    Attributes
    ----------
    stream: :class:`io.IOBase`
        The stream to write the clusters to
    metadata: :class:`dict`
        A set of key-value pairs that describe this
        collection.
    """

    def __init__(self, stream, metadata=None):
        self.stream = stream
        self.metadata = metadata or {}
        self._wrote_metadata = False

    def _write(self, data):
        self.stream.write(data)

    def save(self, cluster):
        """Write ``cluster`` to the output stream, recording its
        members and calculating it's average similarity.

        Parameters
        ----------
        cluster: :class:`SpectrumCluster`
            The spectrum cluster to write out
        """
        if not self._wrote_metadata:
            self.write_metadata()
            self._wrote_metadata = True
        self._save(cluster)

    def _save(self, cluster):
        raise NotImplementedError()

    def save_all(self, clusters):
        """Write each :class:`SpectrumCluster` in ``clusters`` out,
        calling :meth:`save` on each one.

        Parameters
        ----------
        clusters: :class:`collections.Iterable` of :class:`SpectrumCluster`
            The spectrum clusters to write out
        """
        raise NotImplementedError()

    def add_metadata(self, key, value):
        """Add metadata to the writer. That metadata will be flushed
        out upon starting to write clusters out.

        Parameters
        ----------
        key: :class:`str`
            The metadata element's name
        value: :class:`str`, :class:`float`
            The metadata element's value
        """
        if self.wrote_metadata:
            raise TypeError(
                "Cannot add additional metadata, the metadata has already been written")
        self.metadata[key] = value

    def write_metadata(self):
        """Write the accumulated metadata out in a format-appropriate
        manner at the top of the file.
        """
        if self._wrote_metadata:
            raise TypeError("Already wrote metadata!")


class ScanClusterWriter(ScanClusterWriterBase):
    """Writes :class:`ScanCluster` objects to a hierarchical text stream

    Parameters
    ----------
    stream: :class:`io.IOBase`
        The stream to write the clusters to
    """

    def __init__(self, stream, metadata=None):
        super(ScanClusterWriter, self).__init__(stream, metadata)

    def write_metadata(self):
        for key, value in self.metadata.items():
            self._write("#%s = %s\n" % (key, value))
        self._write("\n")

    def _save(self, cluster):
        """Write ``cluster`` as a tab delimited tree, recording its
        members and calculating it's average similarity.

        Parameters
        ----------
        cluster: :class:`SpectrumCluster`
            The spectrum cluster to write out
        """
        self._write("%f\t%d\t%f\n" % (cluster.neutral_mass, len(cluster), cluster.average_similarity()))
        for member in cluster:
            member_source = member.source
            if member_source is not None:
                source_name = member_source.source_file
                if not isinstance(source_name, basestring):
                    if hasattr(source_name, 'name'):
                        source_name = source_name.name
            else:
                source_name = ":detatched:"
            self._write("\t%s\t%s\n" % (source_name, member.id))
        self._write('\n')

    def save_all(self, clusters):
        """Write each :class:`SpectrumCluster` in ``clusters`` out,
        calling :meth:`save` on each one.

        Parameters
        ----------
        clusters: :class:`collections.Iterable` of :class:`SpectrumCluster`
            The spectrum clusters to write out
        """
        for cluster in clusters:
            self.save(cluster)


class JSONScanClusterWriter(ScanClusterWriterBase):
    """A :class:`ScanClusterWriterBase` that uses JSON
    lines.
    """

    def write_metadata(self):
        json.dump(self.metadata, self.stream)
        self._write("\n")

    def _save(self, cluster):
        json.dump(cluster.to_dict(), self.stream)
        self._write("\n")

    def save_all(self, clusters):
        for cluster in clusters:
            self.save(cluster)


class ScanClusterReaderBase(object):
    """Base class for reading spectrum clusters from disk.

    Attributes
    ----------
    stream: :class:`io.IOBase`
        The stream to read the clusters from
    resolver_map: :class:`dict`
        A mapping from scan source name to a :class:`Callable`
        which will return a :class:`~.ScanBase` object representing
        that spectrum.
    metadata: :class:`dict`
        A set of key-value pairs that describe this
        collection.
    clusters: :class:`list`
        The read clusters.
    """
    def __init__(self, stream, resolver_map):
        self.stream = stream
        self.resolver_map = resolver_map
        self.clusters = list()
        self.metadata = {}
        self._generator = None

    def _resolve(self, source, scan_id):
        resolver = self.resolver_map[source]
        try:
            return resolver.get_scan_by_id(scan_id)
        except AttributeError:
            return resolver(scan_id)

    def _parse(self):
        """Parse the cluster collection from :attr:`stream`
        """
        self._load_metadata()
        return self._load_clusters()

    def _load_metadata(self):
        """Read the metadata header from :attr:`stream`.
        """
        raise NotImplementedError()

    def _load_clusters(self):
        """Read the data describing :class:`SpectrumCluster` objects from
        :attr:`stream`.
        """
        raise NotImplementedError()

    def __iter__(self):
        return self

    def __next__(self):
        """Advance the iterator, retrieving the next :class:`SpectrumCluster`

        Returns
        -------
        :class:`SpectrumCluster`
        """
        if self._generator is None:
            self._generator = self._parse()
        return next(self._generator)

    def next(self):
        """Advance the iterator, retrieving the next :class:`SpectrumCluster`

        Returns
        -------
        :class:`SpectrumCluster`
        """
        return self.__next__()


class ScanClusterReader(ScanClusterReaderBase):
    """Reads :class:`SpectrumCluster` objects from hierarchical text files written by
    :class:`ScanClusterWriter`.
    """
    def __init__(self, stream, resolver_map):
        super(ScanClusterReader, self).__init__(stream, resolver_map)
        self._line_buffer = deque()

    def _next_line(self):
        if self._line_buffer:
            return self._line_buffer.popleft()
        return self.stream.readline()

    def _return_line(self, line):
        self._line_buffer.append(line)

    def _stream_lines(self):
        line = self._next_line()
        while line:
            yield line
            line = self._next_line()

    def _load_metadata(self):
        line = self._next_line()
        while line.startswith("#"):
            key, value = line.strip().split(" = ", 1)
            try:
                value = float(value)
            except ValueError:
                value = str(value)
            self.metadata[key] = value
        self._return_line(line)

    def _load_clusters(self):
        current_cluster = []
        mass = None
        similarity = None
        for line in self._stream_lines():
            line = line.rstrip()
            if not line or not line.startswith('\t'):
                if current_cluster:
                    yield SpectrumCluster(current_cluster, mass, similarity)
                    current_cluster = []
                    mass = None
                    similarity = None
                if line:
                    tokens = line.split('\t')
                    mass = float(tokens[0])
                    # size = int(tokens[1])
                    similarity = float(tokens[2])
            elif line.startswith("\t"):
                tokens = line.split("\t")
                source = tokens[1]
                scan_id = tokens[2]
                scan = self._resolve(source, scan_id)
                current_cluster.append(scan)
        if current_cluster:
            yield SpectrumCluster(current_cluster, mass, similarity)


class JSONScanClusterReader(ScanClusterReader):
    """Reads :class:`SpectrumCluster` objects from hierarchical text files written by
    :class:`JSONScanClusterWriter`.
    """
    def _load_metadata(self):
        return json.loads(self.stream.readline())

    def _load_clusters(self):
        line = self.stream.readline()
        while line:
            data = json.loads(line)
            mass = data['neutral_mass']
            similarity = data['average_similarity']
            scans = []
            for scan_bundle in data['scans']:
                scans.append(self._resolve(scan_bundle['source'], scan_bundle['id']))
            yield SpectrumCluster(scans, mass, similarity)
            line = self.stream.readline()



# cython: embedsignature=True

cimport cython
from cpython.ref cimport Py_INCREF
from cpython cimport PyObject
from cpython.float cimport PyFloat_AsDouble
from cpython.list cimport (
    PyList_New, PyList_GET_ITEM, PyList_SET_ITEM,
    PyList_GET_SIZE, PyList_Append, PyList_SetItem)
from cpython.dict cimport PyDict_Next, PyDict_SetItem, PyDict_GetItem

from libc.math cimport floor, fabs
from libc.stdlib cimport malloc, free

from brainpy import PROTON as _PROTON, calculate_mass as _py_calculate_mass
from brainpy._c.isotopic_distribution cimport _isotopic_variants
from brainpy._c.isotopic_distribution cimport TheoreticalPeak
from brainpy._speedup cimport calculate_mass

from ms_peak_picker._c.peak_set cimport FittedPeak


from ms_deisotope.constants import (TRUNCATE_AFTER, IGNORE_BELOW)


IF int == long:
    DEF PY_VERSION = 3
ELSE:
    DEF PY_VERSION = 2
IF UNAME_SYSNAME == "Windows" and PY_VERSION == 2:
    cdef double INFINITY = float('inf')

    cdef int isinf(double x) nogil:
        return fabs(x) == INFINITY
ELSE:
    from libc.math cimport isinf as isinf


@cython.boundscheck(False)
cdef double* _cumulative(TheoreticalIsotopicPattern self):
    cdef:
        size_t i, n
        double total
        double* cumulative_intensities

    n = self.get_size()
    cumulative_intensities = <double*>malloc(sizeof(double) * n)
    total = 0.0
    for i in range(n):
        total += self.get(i).intensity
        cumulative_intensities[i] = total
    return cumulative_intensities


cdef double PROTON
PROTON = _PROTON


cdef inline double _round(double x):
    return floor(x + 0.5)


cdef double neutral_mass(double mz,  int z, double charge_carrier=PROTON):
    return (mz * abs(z)) - (z * charge_carrier)


@cython.cdivision
cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=PROTON):
    return (neutral_mass + (z * charge_carrier)) / abs(z)


cdef void slide(double mz, list peaklist):
    cdef:
        size_t n
        size_t i
        TheoreticalPeak peak, first_peak
        double delta

    n = PyList_GET_SIZE(peaklist) - 1

    first_peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, 0)
    for i in range(n):
        peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i + 1)
        delta = peak.mz - first_peak.mz
        peak.mz = mz + delta
    first_peak.mz = mz
    return


cdef dict scale_dict(dict data, double factor):
    cdef:
        dict scaled
        Py_ssize_t i
        PyObject* key
        PyObject* value

    i = 0
    scaled = {}
    while PyDict_Next(data, &i, &key, &value):
        PyDict_SetItem(scaled, <object>key, _round(PyFloat_AsDouble(<object>value) * factor))
    return scaled


cdef class Averagine(object):
    """An isotopic model which can be used to interpolate the composition
    of a class of molecule given an average monomer composition and a theoretical
    polymer mass

    Implements the :class:`Mapping` interface.

    Attributes
    ----------
    base_composition: Mapping
        A mapping from element symbol to average count (float) of that element
        for the average monomer
    base_mass : float
        The base mass of the average monomer. Calculated from :attr:`base_composition`
    """

    def __init__(self, object base_composition):
        self.base_composition = dict(**base_composition)
        self.base_mass = calculate_mass(self.base_composition)
        assert self.base_mass > 0

    def __reduce__(self):
        return self.__class__, (self.base_composition,), self.__getstate__()

    def __getstate__(self):
        return self.base_composition, self.base_mass

    def __setstate__(self, state):
        composition, mass = state
        self.base_composition = dict(composition)
        self.base_mass = float(mass)

    def __call__(self, double mz, int charge=1, double charge_carrier=PROTON, double truncate_after=0.95,
                 double ignore_below=0.0):
        out = self.isotopic_cluster(mz, charge, charge_carrier, truncate_after, ignore_below)
        return out

    @cython.cdivision
    cpdef dict scale(self, double mz, int charge=1, double charge_carrier=PROTON):
        """Given an m/z and a charge state, interpolate the composition
        of the polymer with the matching neutral mass

        Parameters
        ----------
        mz : float
            The reference m/z to calculate the neutral mass to interpolate from
        charge : int, optional
            The reference charge state to calculate the neutral mass. Defaults to 1
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to the mass of a proton.

        Returns
        -------
        Mapping
            The interpolated composition for the calculated neutral mass,
            rounded to the nearest integer and hydrogen corrected.

        References
        ----------
        Senko, M. W., Beu, S. C., & McLafferty, F. W. (1995). Determination of monoisotopic masses and ion populations
        for large biomolecules from resolved isotopic distributions. Journal of the American Society for Mass
        Spectrometry, 6(4), 229–233. http://doi.org/10.1016/1044-0305(95)00017-8
        """
        cdef:
            double neutral, scale, scaled_mass
            dict scaled
            Py_ssize_t i
            int delta_hydrogen
            int H
            PyObject* key
            PyObject* value
            double count

        neutral = neutral_mass(mz, charge, charge_carrier)

        scale = neutral / self.base_mass
        scaled = scale_dict(self.base_composition, scale)

        delta_hydrogen = <int>_round(calculate_mass(scaled) - neutral)
        H = <int>PyFloat_AsDouble(<object>PyDict_GetItem(scaled, "H"))
        if H > delta_hydrogen:
            PyDict_SetItem(scaled, "H", H - delta_hydrogen)
        else:
            PyDict_SetItem(scaled, "H", 0)

        return scaled

    cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=1, double charge_carrier=PROTON,
                                                      double truncate_after=0.95, double ignore_below=0.0):
        cdef:
            dict composition
            list tid
            TheoreticalIsotopicPattern isotopic_pattern

        composition = self.scale(mz, charge, charge_carrier)
        tid = _isotopic_variants(composition, npeaks=None, charge=charge, charge_carrier=PROTON)

        isotopic_pattern = TheoreticalIsotopicPattern._create(
            tid, (<TheoreticalPeak>PyList_GET_ITEM(tid, 0)).mz, 0)
        isotopic_pattern.shift(mz)

        if truncate_after < 1.0:
            isotopic_pattern.truncate_after(truncate_after)

        if ignore_below > 0:
            isotopic_pattern.ignore_below(ignore_below)

        return isotopic_pattern

    cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=1, double charge_carrier=PROTON,
                                                      double truncate_after=0.95, double ignore_below=0.0):
        """Generate a theoretical isotopic pattern for the given m/z and charge state, thresholded
        by theoretical peak height and density.

        Parameters
        ----------
        mz : float
            The reference m/z to calculate the neutral mass to interpolate from
        charge : int, optional
            The reference charge state to calculate the neutral mass. Defaults to 1
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to the mass of a proton.
        truncate_after : float, optional
            The percentage of the signal in the theoretical isotopic pattern to include.
            Defaults to 0.95, including the first 95% of the signal in the generated pattern
        ignore_below : float, optional
            Omit theoretical peaks whose intensity is below this number.
            Defaults to 0.0

        Returns
        -------
        :class:`.TheoreticalIsotopicPattern`
            The generated and thresholded pattern
        """
        cdef:
            TheoreticalIsotopicPattern out
        out = self._isotopic_cluster(mz, charge, charge_carrier, truncate_after, ignore_below)
        return out

    def __getitem__(self, key):
        return self.base_composition[key]

    def __iter__(self):
        return iter(self.base_composition)

    def keys(self):
        return self.base_composition.keys()

    def values(self):
        return self.base_composition.values()

    def items(self):
        return self.base_composition.items()

    def __repr__(self):
        return "Averagine(%r)" % self.base_composition

    def __richcmp__(self, other, int code):
        if code == 2:
            return self.base_composition == other.base_composition
        elif code == 3:
            return self.base_composition != other.base_composition

    def __hash__(self):
        return hash(frozenset(self.base_composition.items()))


cdef list clone_peak_list(list peaklist):
    cdef:
        size_t i, n
        list result
        TheoreticalPeak peak

    n = PyList_GET_SIZE(peaklist)
    result = PyList_New(n)
    for i in range(n):
        peak = (<TheoreticalPeak>PyList_GET_ITEM(peaklist, i)).clone()
        Py_INCREF(peak)
        PyList_SET_ITEM(result, i, peak)
    return result


cdef double sum_intensity(list peaklist, size_t n):
    cdef:
        size_t i
        double total
        FittedPeak peak
    total = 0
    for i in range(n):
        total += (<FittedPeak>PyList_GET_ITEM(peaklist, i)).intensity
    return total


cdef double top3_scale_factor(TheoreticalIsotopicPattern self, list experimental_distribution):
    cdef:
        double top1, top2, top3, scale
        size_t i, n, top1_index, top2_index, top3_index
        TheoreticalPeak peak
    n = self.get_size()

    top1 = 0
    top2 = 0
    top3 = 0
    top1_index = 0
    top2_index = 0
    top3_index = 0
    for i in range(n):
        peak = self.get(i)
        if peak.intensity > top1:
            top3 = top2
            top3_index = top2_index
            top2 = top1
            top2_index = top1_index
            top1 = peak.intensity
            top1_index = i
        elif peak.intensity > top2:
            top3 = top2
            top3_index = top2_index
            top2 = peak.intensity
            top2_index = i
        elif peak.intensity > top3:
            top3 = peak.intensity
            top3_index = i

    scale = (<FittedPeak>PyList_GET_ITEM(experimental_distribution, top1_index)).intensity / self.get(top1_index).intensity
    scale += (<FittedPeak>PyList_GET_ITEM(experimental_distribution, top2_index)).intensity / self.get(top2_index).intensity
    scale += (<FittedPeak>PyList_GET_ITEM(experimental_distribution, top3_index)).intensity / self.get(top3_index).intensity
    scale /= 3
    return scale



@cython.freelist(1000000)
@cython.final
cdef class TheoreticalIsotopicPattern(object):
    """Represent a theoretical isotopic peak list

    Attributes
    ----------
    peaklist: list of :class:`brainpy.TheoreticalPeak`
        The theoretical isotopic pattern peak list
    origin: float
        The monoisotopic peak's m/z
    """

    @staticmethod
    cdef TheoreticalIsotopicPattern _create(list peaklist, double origin, double offset):
        cdef:
            TheoreticalIsotopicPattern self

        self = TheoreticalIsotopicPattern.__new__(TheoreticalIsotopicPattern)
        self.peaklist = peaklist
        self.origin = origin
        self.offset = offset
        return self

    def __init__(self, peaklist, origin, offset=None):
        self.peaklist = list(peaklist)
        self.origin = origin
        if offset is None:
            offset = self.peaklist[0].mz - origin
        self.offset = offset


    def __getitem__(self, i):
        return self.peaklist[i]

    def __iter__(self):
        return iter(self.peaklist)

    def __len__(self):
        return self.get_size()

    @cython.final
    cdef inline TheoreticalPeak get(self, ssize_t i):
        return <TheoreticalPeak>PyList_GET_ITEM(self.peaklist, i)

    @cython.final
    cdef inline size_t get_size(self):
        return PyList_GET_SIZE(self.peaklist)

    cpdef TheoreticalIsotopicPattern clone(self):
        cdef:
            size_t i, n
            TheoreticalPeak p
            list peaklist
        n = self.get_size()
        peaklist = PyList_New(n)
        for i in range(n):
            p = self.get(i)
            p = TheoreticalPeak._create(p.mz, p.intensity, p.charge)
            Py_INCREF(p)
            PyList_SET_ITEM(peaklist, i, p)
        return TheoreticalIsotopicPattern._create(peaklist, self.origin, self.offset)

    def __reduce__(self):
        return self.__class__, (self.peaklist, self.origin, self.offset)

    @cython.final
    cdef inline list get_processed_peaks(self):
        return self.peaklist

    cdef inline double get_monoisotopic_mz(self):
        return self.origin

    @property
    def monoisotopic_mz(self):
        return self.get_monoisotopic_mz()

    @cython.cdivision
    cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=0.0):
        """Discards peaks whose intensity is below ``ignore_below``.

        After discarding peaks, the pattern will be renormalized to
        sum to ``1.0``

        Parameters
        ----------
        ignore_below : float, optional
            The threshold below which peaks will be discarded

        Returns
        -------
        TheoreticalIsotopicPattern
            self
        """
        cdef:
            double total
            list kept_tid
            size_t i, n
            TheoreticalPeak p
        total = 0
        kept_tid = []
        n = self.get_size()
        for i in range(n):
            p = self.get(i)
            if (p.intensity < ignore_below) and (i > 1):
                continue
            else:
                total += p.intensity
                kept_tid.append(p)
        self.peaklist = kept_tid
        self.offset = self.origin - self.get(0).mz
        n = self.get_size()
        for i in range(n):
            p = self.get(i)
            p.intensity /= total
        return self

    cpdef TheoreticalIsotopicPattern shift(self, double mz):
        """Shift all the m/z of peaks in the isotopic pattern by ``offset``
        m/z.

        This will update :attr:`origin` to reflect the new starting
        monoisotopic m/z.

        Parameters
        ----------
        offset : float
            The amount to shift each peak in the pattern by in m/z

        Returns
        -------
        TheoreticalIsotopicPattern
            self
        """
        cdef:
            TheoreticalPeak peak
            size_t i, n
            double delta

        delta = mz - self.origin
        self.origin = mz

        n = self.get_size()
        for i in range(n):
            peak = self.get(i)
            peak.mz += delta
        return self

    cdef TheoreticalIsotopicPattern clone_shift(self, double mz):
        cdef:
            size_t i, n
            TheoreticalPeak p
            list peaklist
            double delta

        delta = mz - self.origin
        n = self.get_size()
        peaklist = PyList_New(n)
        for i in range(n):
            p = self.get(i)
            p = TheoreticalPeak._create(p.mz + delta, p.intensity, p.charge)
            Py_INCREF(p)
            PyList_SET_ITEM(peaklist, i, p)
        return TheoreticalIsotopicPattern._create(peaklist, mz, self.offset)

    @cython.cdivision
    cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=0.95):
        """Drops peaks from the end of the isotopic pattern
        which make up the last ``1 - truncate_after`` percent
        of the isotopic pattern.

        After truncation, the pattern is renormalized to sum to ``1``

        Parameters
        ----------
        truncate_after : float, optional
            The percentage of the isotopic pattern signal to retain. Defaults
            to 0.95.

        Returns
        -------
        TheoreticalIsotopicPattern
            self
        """
        cdef:
            double cumsum, normalizer
            TheoreticalPeak peak
            list result
            size_t i, n
        cumsum = 0
        result = []
        n = self.get_size()
        for i in range(n):
            peak = self.get(i)
            cumsum += peak.intensity
            PyList_Append(result, peak)
            if cumsum >= truncate_after:
                break
        self.peaklist = result
        n = self.get_size()
        normalizer = 1. / cumsum
        for i in range(n):
            peak = self.get(i)
            peak.intensity *= normalizer
        return self

    cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method="sum"):
        r"""Scales ``self``'s intensity to match the intensity distribution of the
        experimental isotopic pattern in ``experimental_distribution``.

        The ``method`` argument must be one of:

        `"sum"`:
            Scale each peak of the theoretical distribution by the sum of the
            intensity in the experimental distribution such that the sums of their
            intensities are equal.

        `"max"`:
            Select the most abundant peak in the theoretical distribution :math:`t_i`, find it's
            match in the experimental distribution :math:`e_i`, find the scaling factor
            :math:`\alpha = \frac{e_i}{t_i}` which will make :math:`e_i == t_i` and scale all
            peaks in self by :math:`\alpha`

        `"basepeak"`:
            As in `"max"`, except the most abundant peak index is taken from the *experimental*
            distribution

        `"top3"`:
            The as in `"max"`, but the scaling factor is the mean of the scale factors for the
            top three most abundant theoretical peaks.


        Parameters
        ----------
        experimental_distribution : list
            The experimental peaks matched
        method : str, optional
            The scaling method to use. Defaults to ``"sum"``

        Returns
        -------
        TheoreticalIsotopicPattern
            self
        """
        return self._scale(experimental_distribution, method)

    @cython.final
    @cython.cdivision
    cdef inline TheoreticalIsotopicPattern _scale(self, list experimental_distribution, str method="sum"):
        cdef:
            size_t i, j, n
            TheoreticalPeak peak
            FittedPeak expeak
            double total_abundance, maximum, scale_factor
            double weights, scales, w

        n = self.get_size()
        if n == 0:
            raise ValueError("Isotopic Pattern has length 0 (%f, %r)" % (self.origin, self.peaklist))
        if method == "sum":
            total_abundance = sum_intensity(experimental_distribution, n)
            for i in range(n):
                (<TheoreticalPeak>self.get(i)).intensity *= total_abundance
        elif method == "max":
            i = 0
            maximum = 0
            for j in range(n):
                peak = self.get(j)
                if peak.intensity > maximum:
                    maximum = peak.intensity
                    j = i
            scale_factor = (<FittedPeak>PyList_GET_ITEM(
                experimental_distribution, i)).intensity / maximum
            for j in range(n):
                peak = self.get(j)
                peak.intensity *= scale_factor
        elif method == "meanscale":
            scales = 0
            weights = 0
            total_abundance = 0
            for i in range(n):
                expeak = <FittedPeak>PyList_GET_ITEM(experimental_distribution, i)
                peak = self.get(i)
                total_abundance += expeak.intensity
                w = (peak.intensity * expeak.intensity ** 2)
                scales += expeak.intensity / peak.intensity * w
                weights += w
            scale_factor = scales / weights
            for i in range(n):
                peak = self.get(i)
                peak.intensity *= scale_factor
        elif method == 'basepeak':
            i = 1
            j = self.basepeak_index()
            scale_factor =  (<FittedPeak>PyList_GET_ITEM(
                experimental_distribution, j)).intensity / self.get(j).intensity
            if j < n:
                scale_factor += (<FittedPeak>PyList_GET_ITEM(
                                experimental_distribution, j + 1)).intensity / self.get(j + 1).intensity
                i += 1
            if j > 0:
                scale_factor += (<FittedPeak>PyList_GET_ITEM(
                                experimental_distribution, j - 1)).intensity / self.get(j - 1).intensity
                i += 1
            scale_factor /= i
            for i in range(n):
                peak = self.get(i)
                peak.intensity *= scale_factor
        elif method == 'top3':
            scale_factor = top3_scale_factor(self, experimental_distribution)
            for i in range(n):
                peak = self.get(i)
                peak.intensity *= scale_factor
        return self

    cpdef size_t basepeak_index(self):
        cdef:
            size_t i, n, bp_index
            double bp_intensity
            TheoreticalPeak p
        bp_intensity = 0
        bp_index = 0
        for i in range(self.get_size()):
            p = self.get(i)
            if p.intensity > bp_intensity:
                bp_intensity = p.intensity
                bp_index = i
        return bp_index

    @cython.cdivision
    cpdef TheoreticalIsotopicPattern normalize(self):
        cdef:
            size_t i, j, n
            TheoreticalPeak peak
            double total_abundance
        n = self.get_size()
        total_abundance = 0
        for i in range(n):
            total_abundance += self.get(i).intensity
        for i in range(n):
            peak = self.get(i)
            peak.intensity /= total_abundance
        return self

    cpdef TheoreticalIsotopicPattern scale_raw(self, double scale_factor):
        for peak in self:
            peak.intensity *= scale_factor
        return self

    cpdef double drop_last_peak(self):
        """Drop the last peak in the isotopic pattern and re-normalize.

        Returns
        -------
        float:
            The percentage of the total signal in the isotopic pattern remaining
            following the removal of the last peak.
        """
        cdef:
            size_t i, n
            TheoreticalPeak p, tail
            double scaler
            list peaks

        n = self.get_size()
        peaks = PyList_New(n - 1)
        i = 0
        tail = self.get(n - 1)
        scaler = 1 - tail.intensity
        for i in range(n - 1):
            p = self.get(i)
            p.intensity /= scaler
            Py_INCREF(p)
            PyList_SetItem(peaks, i, p)
        self.peaklist = peaks
        return scaler

    cdef TheoreticalIsotopicPattern clone_drop_last(self):
        """Combines the copying traversal with the re-normalization
        of :meth:`drop_last_peak` for efffiency.

        Returns
        -------
        TheoreticalIsotopicPattern:
            A copy of `self` with the last peak removed and renormalized
        """
        cdef:
            size_t i, n
            TheoreticalPeak p, tail
            double scaler
            list peaks

        n = self.get_size()
        peaks = PyList_New(n - 1)
        i = 0
        tail = self.get(n - 1)
        scaler = 1 - tail.intensity
        for i in range(n - 1):
            p = self.get(i).clone()
            p.intensity /= scaler
            Py_INCREF(p)
            PyList_SetItem(peaks, i, p)
        return TheoreticalIsotopicPattern._create(peaks, self.origin, self.offset)

    cpdef double total(self):
        cdef:
            double total
            size_t i, n
            TheoreticalPeak p
        total = 0.0
        n = self.get_size()
        for i in range(n):
            p = self.get(i)
            total += p.intensity
        return total

    def __repr__(self):
        return "TheoreticalIsotopicPattern(%0.4f, charge=%d, (%s))" % (
            self.monoisotopic_mz,
            self.peaklist[0].charge,
            ', '.join("%0.3f" % p.intensity for p in self.peaklist))

    cpdef bint _eq(self, object other):
        cdef:
            list peaklist
            TheoreticalIsotopicPattern other_typed
        if isinstance(other, list):
            peaklist = other
        elif isinstance(other, TheoreticalIsotopicPattern):
            peaklist = other.peaklist
        else:
            raise TypeError(type(other))
        return self.get_processed_peaks() == peaklist

    @cython.final
    cdef bint _eq_inst(self, TheoreticalIsotopicPattern other):
        cdef:
            bint val
            TheoreticalPeak tp1, tp2
            size_t i, n

        n = self.get_size()
        val = n == other.get_size()
        if not val:
            return val
        for i in range(n):
            tp1 = self.get(i)
            tp2 = other.get(i)
            val = tp1._eq(tp2)
            if not val:
                return val
        return val

    def __richcmp__(self, object other, int code):
        if other is None:
            if code == 3:
                return True
            else:
                return False

        if code == 2:
            return self._eq(other)
        elif code == 3:
            return not (self._eq(other))
        else:
            return NotImplemented

    cpdef list incremental_truncation(TheoreticalIsotopicPattern self, double threshold):
        """Create incremental truncations of `self`, dropping the last peak until
        the the total signal in reaches `threshold`

        Parameters
        ----------
        threshold: float
            The minimum percentage of the isotopic pattern to retain.

        Returns
        -------
        :class:`list` of :class:`TheoreticalIsotopicPattern`
        """
        cdef:
            list accumulator
            double* cumulative_intensities
            size_t i, n
            TheoreticalIsotopicPattern template, current

        template = self.clone().normalize()
        n = self.get_size()
        cumulative_intensities = _cumulative(template)
        accumulator = [template]

        i = n - 1
        while i > 0:
            if cumulative_intensities[i - 1] < threshold:
                break
            template = template.clone_drop_last()
            accumulator.append(template)
            i -= 1
        free(cumulative_intensities)
        return accumulator


cdef class AveragineCache(object):
    """A wrapper around a :class:`Averagine` instance which will cache isotopic patterns
    produced for new (m/z, charge) pairs and reuses it for nearby m/z values

    Attributes
    ----------
    averagine : :class:`~Averagine`
        The averagine to use to generate new isotopic patterns
    cache_truncation : float
        Number of decimal places to round off the m/z for caching purposes
    """

    def __init__(self, object averagine, object backend=None, double cache_truncation=1.):
        if backend is None:
            backend = {}
        self.backend = dict(backend)
        if isinstance(averagine, AveragineCache):
            self.averagine = averagine.averagine
            self.cache_truncation = averagine.cache_truncation
            self.backend = averagine.backend.copy()
        else:
            self.averagine = Averagine(averagine)
            self.cache_truncation = cache_truncation
        self.enabled = True

    def __reduce__(self):
        return self.__class__, self.__getstate__()

    def __getstate__(self):
        return self.averagine, self.backend, self.cache_truncation

    def __setstate__(self, state):
        avg, store, trunc = state
        self.averagine = Averagine(avg)
        self.store = dict(store)
        self.cache_truncation = trunc

    @cython.cdivision
    cdef double _make_cache_key(self, double mz):
        cdef double key_mz
        if self.cache_truncation == 0.0:
            key_mz = mz
        else:
            key_mz = _round(mz / self.cache_truncation) * self.cache_truncation
        return key_mz

    def make_cache_key(self, double mz):
        return self._make_cache_key(mz)

    @cython.cdivision
    cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=1, double charge_carrier=PROTON, double truncate_after=0.95,
                                 double ignore_below=0.0):
        cdef:
            double key_mz
            tuple cache_key
            PyObject* pvalue
            TheoreticalIsotopicPattern tid
        if self.enabled:
            key_mz = self._make_cache_key(mz)

            # Attempting to replace this tuple construction (which in turn necessitates packing each
            # numeric argument as a Python object) with a hand-written extension class that can compute
            # its own hash value without invoking any Python operations turns out to be just a bit slower
            # than the bare tuple itself.
            cache_key = (key_mz, charge, charge_carrier, truncate_after)
            pvalue = PyDict_GetItem(self.backend, cache_key)
            if pvalue == NULL:
                tid = self.averagine._isotopic_cluster(key_mz, charge, charge_carrier, truncate_after, ignore_below)
                PyDict_SetItem(self.backend, cache_key, tid.clone())
                tid.shift(mz)
                return tid
            else:
                tid = <TheoreticalIsotopicPattern>pvalue
                tid = tid.clone_shift(mz)
                return tid
        else:
            tid = self.averagine._isotopic_cluster(mz, charge, charge_carrier, truncate_after)
            return tid

    cpdef TheoreticalIsotopicPattern isotopic_cluster(
                                self, double mz, int charge=1, double charge_carrier=PROTON,
                                double truncate_after=0.95, double ignore_below=0.0):
        """Generate a theoretical isotopic pattern for the given m/z and charge state, thresholded
        by theoretical peak height and density.

        Mimics :meth:`.Averagine.isotopic_cluster` but uses the object's cache through
        :meth:`has_mz_charge_pair`.

        Parameters
        ----------
        mz : float
            The reference m/z to calculate the neutral mass to interpolate from
        charge : int, optional
            The reference charge state to calculate the neutral mass. Defaults to 1
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to the mass of a proton.
        truncate_after : float, optional
            The percentage of the signal in the theoretical isotopic pattern to include.
            Defaults to 0.95, including the first 95% of the signal in the generated pattern
        ignore_below : float, optional
            Omit theoretical peaks whose intensity is below this number.
            Defaults to 0.0

        Returns
        -------
        :class:`.TheoreticalIsotopicPattern`
            The generated and thresholded pattern
        """
        return self.has_mz_charge_pair(mz, charge, charge_carrier, truncate_after, ignore_below)

    def __call__(self, double mz, int charge=1, double charge_carrier=PROTON, double truncate_after=0.95,
                 double ignore_below=0.0):
        out = self.isotopic_cluster(mz, charge, charge_carrier, truncate_after, ignore_below)
        return out

    def __getitem__(self, key):
        return self.averagine[key]

    def __iter__(self):
        return iter(self.averagine)

    def keys(self):
        return self.averagine.keys()

    def values(self):
        return self.averagine.values()

    def items(self):
        return self.averagine.items()

    def __repr__(self):
        return "AveragineCache(%r)" % self.averagine

    def clear(self):
        self.backend.clear()

    def populate(self, min_mz=10, max_mz=3005, min_charge=1, max_charge=8, charge_carrier=PROTON,
                 truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
        sign = min_charge / abs(min_charge)
        assert sign == (max_charge / abs(max_charge)), "The polarity of min_charge must match the polarity of max_charge"
        min_charge = abs(min_charge)
        max_charge = abs(max_charge)
        for i in range(int(min_mz), int(max_mz)):
            for j in range(min(max_charge, min_charge), max(min_charge, max_charge) + 1):
                self.isotopic_cluster(
                    i, sign * j, charge_carrier, truncate_after=truncate_after, ignore_below=ignore_below)
        return self

    def __richcmp__(self, other, int code):
        if code == 2:
            return self.averagine == other.averagine and self.backend == other.backend
        elif code == 3:
            return self.averagine != other.averagine and self.backend != other.backend


cdef double _neutron_shift
_neutron_shift = _py_calculate_mass({"C[13]": 1}) - _py_calculate_mass({"C[12]": 1})


@cython.cdivision
cpdef double isotopic_shift(int charge=1):
    return _neutron_shift / <double>(charge)


@cython.cdivision
cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=1800.0, int charge=1):
    cdef:
        double lmbda = mass / lambda_factor
        double p_i = 1.0
        double factorial_acc = 1
        double total = 1.0
        double cur_intensity
        double* intensities
        list result

    intensities = <double*>malloc(sizeof(double) * n_peaks)
    intensities[0] = 1.0
    for i in range(1, n_peaks):
        p_i *= lmbda
        factorial_acc *= i
        cur_intensity = p_i / factorial_acc
        intensities[i] = (cur_intensity if not isinf(cur_intensity) else 0.0)
        total += intensities[i]

    result = []
    iso_shift = isotopic_shift(charge)
    mz = mass_charge_ratio(mass, charge)

    for i in range(n_peaks):
        result.append(TheoreticalPeak._create(mz + i * iso_shift, intensities[i] / total, charge))

    free(intensities)
    return TheoreticalIsotopicPattern(result, origin=mz)



from collections import defaultdict, deque
from ms_deisotope.data_source.scan.mobility_frame import IonMobilityFrame
from ms_deisotope.data_source.scan import ScanBunch, ScanBase

from .scan.scan_iterator import ITERATION_MODE_GROUPED, ITERATION_MODE_SINGLE
from .scan.loader import ScanIterator
from .scan.base import ScanBunch
from .metadata.scan_traits import FAIMS_compensation_voltage


class ScanIteratorProxyBase(object):
    """A wrapper around a :class:`~.RandomAccessScanSource` that alters its behavior
    in some way, either limiting the range it iterates over, or filtering the items
    it would yield.

    This is an abstract base class. It should be extended to be used.

    This class re-uses the same file handle that as :attr:`scan_source`, and should
    not be used interleaved with the original source.

    Attributes
    ----------
    scan_source: :class:`~.RandomAccessScanSource`
        The object to use to load :class:`~.Scan` and :class:`~.ScanBunch` instances.
    iteration_mode: str
        A string denoting :const:`~.ITERATION_MODE_GROUPED` or :const:`~.ITERATION_MODE_SINGLE`
        that controls whether :class:`~.ScanBunch` or :class:`~.Scan` are produced
        by iteration.
    """
    scan_source: ScanIterator

    def __init__(self, scan_source, *args, **kwargs):
        self.scan_source = scan_source

    @property
    def iteration_mode(self):
        """A string denoting :const:`~.ITERATION_MODE_GROUPED` or :const:`~.ITERATION_MODE_SINGLE`
        that controls whether :class:`~.ScanBunch` or :class:`~.Scan` are produced
        by iteration.

        Returns
        -------
        str
        """
        return self.scan_source.iteration_mode

    def has_ms1_scans(self):
        """Checks if this :class:`ScanDataSource` contains MS1 spectra.

        Returns
        -------
        :class:`bool` or :const:`None`
            Returns a boolean value if the presence of MS1 scans is known for certain, or :const:`None`
            if it cannot be determined in the case of missing metadata.
        """
        return self.scan_source.has_ms1_scans()

    def has_msn_scans(self):
        """Checks if this :class:`ScanDataSource` contains MSn spectra.

        Returns
        -------
        :class:`bool` or :const:`None`
            Returns a boolean value if the presence of MSn scans is known for certain, or :const:`None`
            if it cannot be determined in the case of missing metadata.
        """
        return self.scan_source.has_msn_scans()

    def next(self):
        """Advance the iterator, fetching the next :class:`~.ScanBunch` or :class:`~.ScanBase`
        depending upon iteration strategy.

        Returns
        -------
        :class:`~.ScanBunch` or :class:`~.ScanBase`
        """
        raise NotImplementedError()

    def __next__(self):
        """Advance the iterator, fetching the next :class:`~.ScanBunch` or :class:`~.ScanBase`
        depending upon iteration strategy.

        Returns
        -------
        :class:`~.ScanBunch` or :class:`~.ScanBase`
        """
        return self.next()

    def __iter__(self):
        return self


ScanIterator.register(ScanIteratorProxyBase)


class QueryIteratorBase(ScanIteratorProxyBase):
    """A base class for types which iterate over a subset of a :class:`~.RandomAccessScanSource`.

    Attributes
    ----------
    grouped: bool
        Whether or not to produce grouped or single scans.

    """
    def __init__(self, scan_source, grouped=True, *args, **kwargs):
        super(QueryIteratorBase, self).__init__(scan_source, *args, **kwargs)
        self.grouped = grouped

    @property
    def iteration_mode(self):
        if self.grouped:
            return ITERATION_MODE_GROUPED
        else:
            return ITERATION_MODE_SINGLE

    def make_iterator(self, iterator=None, grouped=False):
        """Configure the :class:`ScanIterator`'s behavior, selecting it's iteration strategy over
        either its default iterator or the provided ``iterator`` argument.

        Parameters
        ----------
        iterator : Iterator, optional
            Unused. Included for compatibility with :class:`ScanIterator` API
        grouped : bool, optional
            Whether the iterator should be grouped and produce :class:`.ScanBunch` objects
            or single :class:`.Scan`. If :const:`None` is passed, :meth:`has_ms1_scans` will be
            be used instead. Defaults to :const:`None`.
        """
        self.grouped = grouped
        self._make_iterator(grouped=grouped)

    def _make_iterator(self, grouped=False):
        self.scan_source.reset()
        self.scan_source.make_iterator()

    def reset(self):
        """Reset the iterator, if possible, and clear any caches.

        Resets the underlying :class:`~.RandomAccessScanSource`
        """
        self.scan_source.reset()
        self.make_iterator(grouped=self.grouped)


class TimeIntervalIterator(QueryIteratorBase):
    """Query over a retention time interval.

    Attributes
    ----------
    start: float
        The time to start the query iterator from
    end: float
        The time to end the query iterator at

    """
    def __init__(self, scan_source, start=None, end=None, grouped=True, *args, **kwargs):
        super(TimeIntervalIterator, self).__init__(
            scan_source, grouped, * args, **kwargs)
        self.start = start
        self.end = end
        self._update_bounds()
        self.make_iterator(grouped=grouped)

    def _update_bounds(self):
        if self.start is None:
            try:
                first_entry = self.scan_source[0]
                self.start = first_entry.scan_time
            except (AttributeError, TypeError):
                self.start = 0
        if self.end is None:
            try:
                last_entry = self.scan_source[-1]
                self.end = last_entry.scan_time
            except (AttributeError, TypeError):
                self.end = float('inf')

    def _make_iterator(self, grouped=False):
        self.scan_source.start_from_scan(rt=self.start, grouped=self.grouped)

    def next(self):
        result = next(self.scan_source)
        if self.grouped:
            if result.precursor.scan_time > self.end:
                raise StopIteration()
            return result
        else:
            if result.scan_time > self.end:
                raise StopIteration()
            return result


class IndexIntervalIterator(QueryIteratorBase):
    """Query over a scan index interval.

    Attributes
    ----------
    start: int
        The index to start the query iterator from
    end: int
        The index to end the query iterator at

    """

    def __init__(self, scan_source, start=None, end=None, grouped=True, *args, **kwargs):
        super(IndexIntervalIterator, self).__init__(
            scan_source, grouped, * args, **kwargs)
        self.start = start
        self.end = end
        self._update_bounds()
        self.make_iterator(grouped=grouped)

    def _update_bounds(self):
        if self.start is None:
            try:
                first_entry = self.scan_source[0]
                self.start = first_entry.index
            except (AttributeError, TypeError):
                self.start = 0
        if self.end is None:
            try:
                last_entry = self.scan_source[-1]
                self.end = last_entry.index
            except (AttributeError, TypeError):
                self.end = float('inf')

    def _make_iterator(self, grouped=False):
        self.scan_source.start_from_scan(index=self.start, grouped=self.grouped)

    def next(self):
        result = next(self.scan_source)
        if self.grouped:
            if result.precursor.index > self.end:
                raise StopIteration()
            return result
        else:
            if result.index > self.end:
                raise StopIteration()
            return result


def scan_range(scan_source, time=None, index=None, grouped=True, *args, **kwargs):
    """Create an iterator proxy over `scan_source` spanning a specified range in time
    or index.

    If neither `time` nor `index` is provided, `scan_source` is returned unchanged.
    If both are provided, an error is thrown.

    Parameters
    ----------
    scan_source : :class:`~.RandomAccessScanSource`
        The scan source to iterate over
    time : tuple[float, float], optional
        The start and stop times to use
    index : tuple[int, int], optional
        The start and top scan indices to use
    grouped : bool, optional
        Whether or not to create a :class:`~.ScanBunch` iterator or a :class:`~.Scan` iterator
    *args:
    **kwargs:
        Forwarded to the iterator proxy.

    Returns
    -------
    :class:`ScanIterator`

    Raises
    ------
    ValueError:
        If both `time` and `index` are provided.
    """
    if time and index:
        raise ValueError("Only one of time and index intervals may be specified")
    if time:
        return TimeIntervalIterator(
            scan_source, time[0], time[1], grouped=grouped, *args, **kwargs)
    elif index:
        return IndexIntervalIterator(
            scan_source, index[0], index[1], grouped=grouped, *args, **kwargs)
    else:
        return scan_source


class ScanIteratorFilterBase(ScanIteratorProxyBase):
    """A base class for types which filter or transform scans produced
    by a :class:`~.ScanIterator` or :class:`~.RandomAccessScanSource`.

    """
    def __init__(self, scan_source, *args, **kwargs):
        super(ScanIteratorFilterBase, self).__init__(scan_source, *args, **kwargs)
        self._generator = None

    def filter_scan_bunch(self, scan_group, **kwargs):
        """Filter a scan bunch

        Parameters
        ----------
        scan : :class:`~.ScanBunch`
            The scan bunch to filter

        Yields
        ------
        :class:`~.ScanBunch`
        """
        raise NotImplementedError()

    def filter_scan(self, scan, **kwargs):
        """Filter a single scan

        Parameters
        ----------
        scan : :class:`~.Scan`
            The single scan to filter

        Yields
        ------
        :class:`~.Scan`
        """
        raise NotImplementedError()

    def filter(self, result, **kwargs):
        """Filter scans, producing only those that passed the criteria.

        This returns a generator which yields the same type as `result`

        This method will dispatch to :meth:`filter_scan` or :meth:`filter_scan_bunch`,
        whichever is appropriate for `result`

        Parameters
        ----------
        result : :class:`~.Scan` or :class:`~.ScanBunch`
            The scan or bunch to filter.

        Yields
        ------
        :class:`~.Scan` or :class:`~.ScanBunch`
        """
        if self.scan_source.iteration_mode == ITERATION_MODE_GROUPED:
            return self.filter_scan_bunch(result, **kwargs)
        else:
            return self.filter_scan(result, **kwargs)

    def _generate(self):
        for batch in self.scan_source:
            for result in self.filter(batch):
                yield result

    def next(self):
        if self._generator is None:
            self._generator = self._generate()
        return next(self._generator)


class _PredicateFilterIterator(ScanIteratorFilterBase):
    def _check(self, scan):
        raise NotImplementedError()

    def filter_scan_bunch(self, scan_group, **kwargs):
        if scan_group.products:
            yield scan_group.__class__(
                scan_group.precursor if self._check(
                    scan_group.precursor) else None,
                [scan for scan in scan_group.products if self._check(scan)])

    def filter_scan(self, scan, **kwargs):
        if self._check(scan):
            yield scan

    @property
    def key(self):
        raise NotImplementedError()


class PolarityFilter(_PredicateFilterIterator):
    """A Scan Iterator that filters out scans with a polarity
    which do not match :attr:`polarity`

    Attributes
    ----------
    polarity: int
        The polarity to match

    """

    def __init__(self, scan_source, polarity, *args, **kwargs):
        super(PolarityFilter, self).__init__(scan_source, *args, **kwargs)
        self.polarity = polarity

    def _check(self, scan):
        return scan.polarity == self.polarity

    @property
    def key(self):
        return self.polarity


class MSLevelFilter(_PredicateFilterIterator):
    """A Scan Iterator that filters out scans with MS levels
    which do not match :attr:`ms_level`

    Attributes
    ----------
    ms_level: int
        The MS level to require matching

    """
    def __init__(self, scan_source, ms_level, *args, **kwargs):
        super(MSLevelFilter, self).__init__(scan_source, *args, **kwargs)
        self.ms_level = ms_level

    def _check(self, scan):
        return scan.ms_level == self.ms_level

    @property
    def key(self):
        return self.ms_level


class MassAnalyzerFilter(_PredicateFilterIterator):
    """A Scan Iterator that filters out scans which were not performed using
    :attr:`mass_analyzer`

    Attributes
    ----------
    mass_analyzer: :class:`~.Component`
        The mass analyzer to match
    """
    def __init__(self, scan_source, mass_analyzer, *args, **kwargs):
        super(MassAnalyzerFilter, self).__init__(scan_source, *args, **kwargs)
        from .metadata.instrument_components import analyzer_types, Component
        if not isinstance(mass_analyzer, Component):
            mass_analyzer = analyzer_types[mass_analyzer]
        self.mass_analyzer = mass_analyzer

    def _check(self, scan):
        ic = scan.instrument_configuration
        if ic is None:
            return False
        for analyzer in ic.analyzers:
            if analyzer.is_a(self.mass_analyzer):
                return True
        return False

    @property
    def key(self):
        return self.mass_analyzer


class CallableFilter(_PredicateFilterIterator):
    """A Scan Iterator that filters out scans which do not pass
    a user-provided callable

    Attributes
    ----------
    filter_fn: :class:`Callable`
        The callable object to use to test if a scan should be used
    """

    def __init__(self, scan_source, filter_fn, *args, **kwargs):
        super(CallableFilter, self).__init__(scan_source, *args, **kwargs)
        self.filter_fn = filter_fn

    def _check(self, scan):
        return self.filter_fn(scan)

    @property
    def key(self):
        return self.filter_fn


filter_scans = CallableFilter


class HasProductMZFilter(_PredicateFilterIterator):
    product_mz: float
    error_tolerance: float

    def __init__(self, scan_source, product_mz, error_tolerance, *args, **kwargs):
        super(HasProductMZFilter, self).__init__(scan_source, *args, **kwargs)
        self.product_mz = product_mz
        self.error_tolerance = error_tolerance

    def _check(self, scan: ScanBase):
        if scan.peak_set is None and scan.deconvoluted_peak_set is None:
            scan.pick_peaks()
        if scan.peak_set is not None:
            return bool(scan.peak_set.all_peaks_for(self.product_mz, self.error_tolerance))
        elif scan.deconvoluted_peak_set is not None:
            lower = self.product_mz - self.product_mz * self.error_tolerance
            upper = self.product_mz + self.product_mz * self.error_tolerance
            return bool(scan.deconvoluted_peak_set.between(lower, upper, use_mz=True))
        raise ValueError("Could not find peak list to pick from!")

    @property
    def key(self):
        return self.product_mz


class HasPrecursorLossMZFilter(_PredicateFilterIterator):
    loss_mz: float
    error_tolerance: float

    def __init__(self, scan_source, loss_mz, error_tolerance, *args, **kwargs):
        super(HasPrecursorLossMZFilter, self).__init__(
            scan_source, *args, **kwargs)
        self.loss_mz = loss_mz
        self.error_tolerance = error_tolerance

    def _check(self, scan: ScanBase):
        if not scan.precursor_information:
            return False
        mz = scan.precursor_information.mz - self.loss_mz
        if scan.peak_set is None and scan.deconvoluted_peak_set is None:
            scan.pick_peaks()
        if scan.peak_set is not None:
            return bool(scan.peak_set.all_peaks_for(mz, self.error_tolerance))
        elif scan.deconvoluted_peak_set is not None:
            lower = mz - mz * self.error_tolerance
            upper = mz + mz * self.error_tolerance
            return bool(scan.deconvoluted_peak_set.between(lower, upper, use_mz=True))
        raise ValueError("Could not find peak list to pick from!")

    @property
    def key(self):
        return self.loss_mz


class MS1MergingTransformer(ScanIteratorProxyBase):

    def __init__(self, scan_source, ms1_bin_width=10, mz_resolution=None, *args, **kwargs):
        super(MS1MergingTransformer, self).__init__(scan_source, *args, **kwargs)
        self._generator = None
        self.ms1_bin_width = ms1_bin_width
        self.mz_resolution = mz_resolution
        self.ms1_buffer = []
        self.msn_buffer = []

    def _redirect_precursor_ids(self, products, precursor):
        level_to_redirect = precursor.ms_level + 1
        # Don't re-direct MS3+ scans
        for product in products:
            if product.ms_level == level_to_redirect:
                product.precursor_information.precursor_scan_id = precursor.id
        return products

    def _generate(self):
        half_width = self.ms1_bin_width // 2
        for batch in self.scan_source:
            if len(self.ms1_buffer) == self.ms1_bin_width:
                ref_scan = self.ms1_buffer[0]
                merged_scan = ref_scan.average_with(self.ms1_buffer[1:], self.mz_resolution)
                # Scale up to be the sum not the mean
                merged_scan.arrays *= self.ms1_bin_width
                bunch = ScanBunch(merged_scan, self._redirect_precursor_ids(self.msn_buffer, merged_scan))
                yield bunch
                self.msn_buffer = []
                self.ms1_buffer = self.ms1_buffer[half_width:]

            self.ms1_buffer.append(batch.precursor)
            self.msn_buffer.extend(batch.products)

        if self.ms1_buffer:
            ref_scan = self.ms1_buffer[0]
            merged_scan = ref_scan.average_with(self.ms1_buffer[1:], self.mz_resolution)
            # Scale up to be the sum not the mean
            merged_scan.arrays *= self.ms1_bin_width
            bunch = ScanBunch(merged_scan, self._redirect_precursor_ids(self.msn_buffer, merged_scan))
            yield bunch


    def next(self):
        if self._generator is None:
            self._generator = self._generate()
        return next(self._generator)


class TimeOrderMergingIterator(object):
    def __init__(self, sources):
        self.sources = list(sources)
        self._init_heads()

    def _init_heads(self):
        n = len(self.sources)
        self.heads = [None] * n
        for i in range(n):
            self._advance(i)

    def __next__(self):
        result = self.get_next_head()
        if result is None:
            raise StopIteration()
        return result

    next = __next__

    def __iter__(self):
        return self

    def count_exhausted(self):
        n = 0
        for h in self.heads:
            n += h is None
        return n

    def _advance(self, i):
        try:
            self.heads[i] = next(self.sources[i])
        except StopIteration:
            self.heads[i] = None

    def get_next_head(self):
        best = None
        best_i = None
        best_time = float('inf')
        for i, head in enumerate(self.heads):
            if head is None:
                continue
            time = self.get_time(head)
            assert time is not None
            if time < best_time:
                best = head
                best_i = i
                best_time = time
        if best is None:
            return best
        i = best_i
        self._advance(i)
        return best

    def get_time(self, obj):
        if isinstance(obj, ScanBunch):
            if obj.precursor is not None:
                return self.get_time(obj.precursor)
            elif obj.products:
                return self.get_time(obj.products[0])
        elif isinstance(obj, ScanBase):
            return obj.scan_time
        elif isinstance(obj, IonMobilityFrame):
            return obj.time
        else:
            # Just guessing at this point
            return obj.scan_time


class FAIMSFilter(_PredicateFilterIterator):
    def __init__(self, scan_source, compensation_voltage, *args, **kwargs):
        super(FAIMSFilter, self).__init__(scan_source, *args, **kwargs)
        self.compensation_voltage = compensation_voltage

    def _check(self, scan):
        if scan.has_ion_mobility():
            if scan.ion_mobility_type == FAIMS_compensation_voltage:
                if scan.drift_time == self.compensation_voltage:
                    return scan

    @property
    def key(self):
        return self.compensation_voltage


class QueueIterator(ScanIteratorProxyBase):
    def __init__(self, scan_source, data=None):
        self.scan_source = scan_source
        self.data = deque(data or [])

    def __next__(self):
        return self.data.popleft()

    def has_value(self):
        return bool(self.data)

    def _feed(self, value):
        self.data.append(value)

    def __len__(self):
        return len(self.data)


class DemultiplexingIteratorBase(ScanIteratorProxyBase):
    def __init__(self, scan_source, buffer_size=10, *args, **kwargs):
        super(DemultiplexingIteratorBase, self).__init__(
            scan_source, *args, **kwargs)
        self.channels = dict()
        self.buffer_size = buffer_size
        self.has_more = self.feed()

    @property
    def iteration_mode(self):
        return ITERATION_MODE_SINGLE

    def _new_channel(self, scan):
        raise NotImplementedError()

    def __getitem__(self, key):
        return self.channels[key]

    def __iter__(self):
        return self

    def _get_channel_name(self, scan_filter):
        return scan_filter.key

    def feed(self):
        loading = sum([len(v[1]) for v in self.channels.values()])
        while loading < self.buffer_size:
            try:
                batch = next(self.scan_source)
            except StopIteration:
                return False
            if self.scan_source.iteration_mode == ITERATION_MODE_SINGLE:
                for channel_name, (scan_filter, queue) in self.channels.items():
                    if scan_filter._check(batch):
                        queue._feed(batch)
                        break
                else:
                    self._new_channel(batch)
            else:
                for scan in [batch.precursor] + batch.products:
                    for channel_name, (scan_filter, queue) in self.channels.items():
                        if scan_filter._check(scan):
                            queue._feed(scan)
                            break
                    else:
                        self._new_channel(scan)
            loading = sum([len(v[1]) for v in self.channels.values()])
        return True

    def __next__(self):
        if not self.has_more:
            raise StopIteration()
        meta = {}
        for channel_name, (scan_filter, queue) in self.channels.items():
            if queue.has_value():
                meta[channel_name] = next(scan_filter)
            else:
                meta[channel_name] = None
        self.has_more = self.feed()
        return meta


class FAIMSDemultiplexingIterator(DemultiplexingIteratorBase):
    def __init__(self, scan_source, buffer_size=10, *args, **kwargs):
        super(FAIMSDemultiplexingIterator, self).__init__(scan_source, buffer_size, *args, **kwargs)

    def _new_channel(self, scan):
        queue = QueueIterator(self)
        scan_filter = FAIMSFilter(queue, scan.drift_time)
        queue._feed(scan)
        self.channels[self._get_channel_name(scan_filter)] = (scan_filter, queue)




import os
import logging
import sys
import multiprocessing
import traceback
import pickle

from typing import Any, Callable, Deque, Dict, List, Optional, Tuple, Union
from collections import deque
from multiprocessing import Process
from queue import Empty as QueueEmpty

from ms_deisotope.data_source import Scan, RandomAccessScanSource, ScanIterator


import ms_deisotope
from ms_deisotope.data_source.scan.scan import ProcessedScan

from ms_deisotope.processor import (
    ScanProcessor, MSFileLoader,
    NoIsotopicClustersError, EmptyScanError)

from ms_deisotope.task import show_message

try:
    from pyzstd import decompress, compress
except ImportError:
    from gzip import decompress, compress


class CompressedPickleMessage(object):
    def __init__(self, obj=None):
        self.obj = obj

    def __getstate__(self):
        state = {
            "payload": compress(pickle.dumps(self.obj, -1))
        }
        return state

    def __setstate__(self, state):
        self.obj = pickle.loads(decompress(state['payload']))

    def __reduce__(self):
        return self.__class__, (None, ), self.__getstate__()

    def __iter__(self):
        yield self.obj


DONE = b"--NO-MORE--"
SCAN_STATUS_GOOD = b"good"
SCAN_STATUS_SKIP = b"skip"


class ScanTransmissionMixin(object):
    output_queue: multiprocessing.JoinableQueue

    def skip_entry(self, index: int, ms_level: int):
        self.output_queue.put((SCAN_STATUS_SKIP, index, ms_level))

    def skip_scan(self, scan: Scan):
        self.output_queue.put((SCAN_STATUS_SKIP, scan.index, scan.ms_level))

    def send_scan(self, scan: Scan):
        scan = scan.pack()
        # this attribute is not needed, and for MS1 scans is dangerous
        # to pickle.
        # It can pull other scans which may not yet have been packed
        # into the message sent back to the main process which in
        # turn can form a reference cycle and eat a lot of memory
        scan.product_scans = []
        if isinstance(scan, ProcessedScan):
            if scan.deconvoluted_peak_set is not None:
                scan.peak_set = None
                for peak in scan.deconvoluted_peak_set:
                    if peak.fit is not None:
                        peak.fit = None
        self.output_queue.put(
            (CompressedPickleMessage(scan), scan.index, scan.ms_level))


class _ProcessHelper:
    def try_set_process_name(self, name=None):
        """This helper method may be used to try to change a process's name
        in order to make discriminating which role a particular process is
        fulfilling. This uses a third-party utility library that may not behave
        the same way on all platforms, and therefore this is done for convenience
        only.

        Parameters
        ----------
        name : str, optional
            A name to set. If not provided, will check the attribute ``process_name``
            for a non-null value, or else have no effect.
        """
        if name is None:
            name = getattr(self, 'process_name', None)
        if name is None:
            return
        try:
            import setproctitle
            setproctitle.setproctitle(name)
        except (ImportError, AttributeError):
            pass


class ScanIDYieldingProcess(Process, ScanTransmissionMixin, _ProcessHelper):
    ms_file_path: os.PathLike
    scan_id_queue: multiprocessing.JoinableQueue
    loader: Union[ScanIterator, RandomAccessScanSource]

    start_scan: str
    end_scan: str
    max_scans: Optional[int]
    end_scan_index: Optional[int]

    ignore_tandem_scans: bool
    batch_size: int

    no_more_event: Optional[multiprocessing.Event]
    log_handler: Callable

    def __init__(self, ms_file_path: os.PathLike, scan_id_queue: multiprocessing.JoinableQueue, start_scan: str = None,
                 max_scans: Optional[int]=None, end_scan: str=None, no_more_event: Optional[multiprocessing.Event]=None,
                 ignore_tandem_scans: bool=False, batch_size: int=1, log_handler: Callable=None,
                 output_queue: Optional[multiprocessing.JoinableQueue]=None):
        if log_handler is None:
            log_handler = show_message
        Process.__init__(self)
        self.daemon = True
        self.ms_file_path = ms_file_path
        self.scan_id_queue = scan_id_queue
        self.loader = None

        self.start_scan = start_scan
        self.max_scans = max_scans
        self.end_scan = end_scan
        self.end_scan_index = None
        self.passed_first_batch = False
        self.ignore_tandem_scans = ignore_tandem_scans
        self.batch_size = batch_size

        self.log_handler = log_handler

        self.no_more_event = no_more_event
        self.output_queue = output_queue

    def _make_scan_batch(self) -> Tuple[
            List[Tuple[str, List[str]]],
            List[str]
        ]:
        batch = []
        scan_ids = []
        for _ in range(self.batch_size):
            try:
                bunch = next(self._iterator)
                scan, products = bunch
                if scan is not None:
                    scan_id = scan.id
                    if scan.index > self.end_scan_index:
                        break
                else:
                    scan_id = None
                products = [prod for prod in products if prod.index <= self.end_scan_index]
                product_scan_ids = [p.id for p in products]
            except StopIteration:
                break
            except Exception as e:
                self.log_handler("An error occurred in _make_scan_batch", e)
                break
            if not self.ignore_tandem_scans:
                batch.append((scan_id, product_scan_ids, True))
            else:
                batch.append((scan_id, product_scan_ids, False))
            scan_ids.append(scan_id)
        return batch, scan_ids

    def _initialize_iterator(self):
        if self.start_scan is not None:
            try:
                self.loader.start_from_scan(
                    self.start_scan, require_ms1=self.loader.has_ms1_scans(), grouped=True)
            except IndexError as e:
                self.log_handler(
                    "An error occurred while locating start scan", e)
                self.loader.reset()
                self.loader.make_iterator(grouped=True)
            except AttributeError as e:
                self.log_handler(
                    "The reader does not support random access, start time will be ignored", e)
                self.loader.reset()
                self.loader.make_iterator(grouped=True)
        else:
            self.loader.make_iterator(grouped=True)
        self._iterator = self.loader

    def _prepare_end_scan_marker(self) -> Optional[str]:
        end_scan = self.end_scan
        if end_scan is None:
            try:
                self.end_scan_index = len(self.loader)
            except AttributeError:
                self.end_scan_index = sys.maxint
        else:
            self.end_scan_index = self.loader.get_scan_by_id(end_scan).index
        return end_scan

    def _open_ms_file(self) -> Union[ScanIterator, RandomAccessScanSource]:
        self.loader = MSFileLoader(self.ms_file_path, decode_binary=False)
        return self.loader

    def run(self):
        self.try_set_process_name("ms-deisotope-sched")
        self._open_ms_file()
        self._initialize_iterator()

        count: int = 0
        last: int = 0
        if self.max_scans is None:
            max_scans = float('inf')
        else:
            max_scans = self.max_scans

        end_scan = self._prepare_end_scan_marker()
        while count < max_scans:
            try:
                batch, ids = self._make_scan_batch()
                if len(batch) > 0:
                    self.scan_id_queue.put(batch)
                count += len(ids)
                if (count - last) > 1000:
                    last = count
                    self.scan_id_queue.join()
                if (end_scan in ids and end_scan is not None) or len(ids) == 0:
                    self.log_handler("End Scan Found")
                    break
            except StopIteration:
                break
            except Exception as e:
                self.log_handler("An error occurred while fetching scans", e)
                break

        if self.no_more_event is not None:
            self.no_more_event.set()
            self.log_handler("All Scan IDs have been dealt. %d scan bunches." % (count,))
        else:
            self.scan_id_queue.put(DONE)


class ScanTransformMixin(object):
    _batch_store: Deque[Tuple[str, List[str], bool]]

    input_queue: multiprocessing.JoinableQueue

    def log_error(self, error: Exception, scan_id: str, scan: Scan, product_scan_ids: List[str]):
        tb = traceback.format_exc()
        self.log_handler(
            "An %r occurred for %s (index %r) in Process %r\n%s" % (
                error, scan_id, scan.index, multiprocessing.current_process(),
                tb))

    def _init_batch_store(self):
        self._batch_store = deque()

    def get_work(self, block: bool=True, timeout: float=30) -> Tuple[str, List[str], bool]:
        if self._batch_store:
            return self._batch_store.popleft()
        else:
            batch = self.input_queue.get(block, timeout)
            self.input_queue.task_done()
            self._batch_store.extend(batch)
            result = self._batch_store.popleft()
            return result

    def log_message(self, message):
        self.log_handler(message + ", %r" %
                         (multiprocessing.current_process().name))

    def all_work_done(self) -> bool:
        return self._work_complete.is_set()

    def make_scan_transformer(self, loader=None):
        raise NotImplementedError()

    _loggers_to_silence = ["ms_deisotope.scan_processor"]

    def _silence_loggers(self):
        nologs = self._loggers_to_silence
        if not self.deconvolute:
            nologs.append("deconvolution")

        debug_mode = os.getenv("MS_DEISOTOPE_DEBUG")
        if debug_mode:
            handler = logging.FileHandler(
                "ms-deisotope-deconvolution-debug-%s.log" % (os.getpid()), 'w')
            fmt = logging.Formatter(
                "%(asctime)s - %(name)s:%(filename)s:%(lineno)-4d - %(levelname)s - %(message)s",
                "%H:%M:%S")
            handler.setFormatter(fmt)
        for logname in nologs:
            logger_to_silence = logging.getLogger(logname)
            if debug_mode:
                logger_to_silence.setLevel("DEBUG")
                logger_to_silence.addHandler(handler)
            else:
                logger_to_silence.propagate = False
                logger_to_silence.setLevel("CRITICAL")
                logger_to_silence.addHandler(logging.NullHandler())


class ScanBunchLoader(object):
    queue: Deque[Tuple[str, List[str]]]
    loader: Union[RandomAccessScanSource, ScanIterator]

    def __init__(self, mzml_loader):
        self.loader = mzml_loader
        self.queue = deque()

    def put(self, scan_id: str, product_scan_ids: List[str]):
        self.queue.append((scan_id, product_scan_ids))

    def get(self) -> Tuple[Scan, List[Scan]]:
        scan_id, product_scan_ids = self.queue.popleft()
        if scan_id is not None:
            precursor = self.loader.get_scan_by_id(scan_id)
        else:
            precursor = None
        products = [self.loader.get_scan_by_id(
            pid) for pid in product_scan_ids if pid is not None]
        if precursor:
            precursor.product_scans = products
        return (precursor, products)


class DeconvolutingScanTransformingProcess(Process, ScanTransformMixin, ScanTransmissionMixin, _ProcessHelper):
    """DeconvolutingScanTransformingProcess describes a child process that consumes scan id bunches
    from a shared input queue, retrieves the relevant scans, and preprocesses them using an
    instance of :class:`ms_deisotope.processor.ScanProcessor`, sending the reduced result
    to a shared output queue.

    Attributes
    ----------
    input_queue : multiprocessing.JoinableQueue
        A shared input queue which contains payloads of bunches of
        scan ids
    ms1_deconvolution_args : dict
        Parameters passed to :class:`ms_deisotope.processor.ScanProcessor`
    ms1_peak_picking_args : dict
        Parameters passed to :class:`ms_deisotope.processor.ScanProcessor`
    msn_deconvolution_args : dict
        Parameters passed to :class:`ms_deisotope.processor.ScanProcessor`
    msn_peak_picking_args : dict
        Parameters passed to :class:`ms_deisotope.processor.ScanProcessor`
    ms_file_path : str
        Path to the spectral data file on disk
    no_more_event : multiprocessing.Event
        An event which will be set when the process feeding the input
        queue has run out of items to add, indicating that any QueueEmptyException
        should be treated as a signal to finish rather than to wait for
        new input
    output_queue : multiprocessing.JoinableQueue
        A shared output queue which this object will put
        :class:`ms_deisotope.data_source.common.ProcessedScan` bunches onto.
    """

    ms_file_path: os.PathLike
    input_queue: multiprocessing.Queue
    output_queue: multiprocessing.Queue

    loader: Union[ScanIterator, RandomAccessScanSource]

    ms1_averaging: int
    envelope_selector: Callable
    transformer: Optional[ScanProcessor]

    ms1_peak_picking_args: Dict[str, Any]
    msn_peak_picking_args: Dict[str, Any]

    ms1_deconvolution_args: Dict[str, Any]
    msn_deconvolution_args: Dict[str, Any]

    too_many_peaks_threshold: int
    default_precursor_ion_selection_window: float
    deconvolute: bool

    _work_complete: multiprocessing.Event
    no_more_event: Optional[multiprocessing.Event]

    log_handler: Callable

    def __init__(self, ms_file_path, input_queue, output_queue,
                 no_more_event=None, ms1_peak_picking_args=None,
                 msn_peak_picking_args=None,
                 ms1_deconvolution_args=None, msn_deconvolution_args=None,
                 envelope_selector=None, ms1_averaging=0, log_handler=None,
                 deconvolute=True, verbose=False, too_many_peaks_threshold=7000,
                 default_precursor_ion_selection_window=1.5):
        if log_handler is None:
            log_handler = show_message

        if ms1_peak_picking_args is None:
            ms1_peak_picking_args = {
                "transforms": [],
                "start_mz": 250
            }
        if msn_peak_picking_args is None:
            msn_peak_picking_args = {
                "transforms": []
            }
        if ms1_deconvolution_args is None:
            ms1_deconvolution_args = {
                "scorer": ms_deisotope.scoring.PenalizedMSDeconVFitter(20., 2),
                "charge_range": (1, 8),
                "averagine": ms_deisotope.peptide,
                "use_quick_charge": True,
            }
        if msn_deconvolution_args is None:
            msn_deconvolution_args = {
                "scorer": ms_deisotope.scoring.MSDeconVFitter(10.),
                "charge_range": (1, 8),
                "averagine": ms_deisotope.peptide,
                "use_quick_charge": True,
            }

        Process.__init__(self)
        self.daemon = True
        self.verbose = verbose
        self._init_batch_store()
        self.ms_file_path = ms_file_path
        self.input_queue = input_queue
        self.output_queue = output_queue

        self.ms1_peak_picking_args = ms1_peak_picking_args
        self.msn_peak_picking_args = msn_peak_picking_args
        self.ms1_deconvolution_args = ms1_deconvolution_args
        self.msn_deconvolution_args = msn_deconvolution_args
        self.envelope_selector = envelope_selector
        self.ms1_averaging = ms1_averaging
        self.deconvolute = deconvolute

        self.transformer = None

        self.no_more_event = no_more_event
        self._work_complete = multiprocessing.Event()
        self.log_handler = log_handler
        self.too_many_peaks_threshold = too_many_peaks_threshold
        self.default_precursor_ion_selection_window = default_precursor_ion_selection_window

    def make_scan_transformer(self, loader: Union[ScanIterator, RandomAccessScanSource] = None) -> ScanProcessor:
        self.transformer = ScanProcessor(
            loader,
            ms1_peak_picking_args=self.ms1_peak_picking_args,
            msn_peak_picking_args=self.msn_peak_picking_args,
            ms1_deconvolution_args=self.ms1_deconvolution_args,
            msn_deconvolution_args=self.msn_deconvolution_args,
            loader_type=lambda x: x,
            envelope_selector=self.envelope_selector,
            ms1_averaging=self.ms1_averaging,
            default_precursor_ion_selection_window=self.default_precursor_ion_selection_window)
        return self.transformer

    def _process_ms1(self, scan, product_scans) -> Tuple[Scan, List, List[Scan]]:
        scan, priorities, product_scans = self.transformer.process_scan_group(
            scan, product_scans)
        return scan, priorities, product_scans

    def _deconvolute_ms1(self, scan: Scan, priorities: List, product_scans: List[Scan]):
        self.transformer.deconvolute_precursor_scan(scan, priorities, product_scans)

    def _handle_ms1_scan(self, scan: Scan, product_scans: List[Scan], scan_id: str, product_scan_ids: List[str]) -> Tuple[Scan, List[Scan]]:
        try:
            # Check if the m/z array is empty, if so skip the scan. This may trigger
            # arbitrary data loading and decompression behavior, so it can fail.
            if len(scan.arrays[0]) == 0:
                self.skip_scan(scan)
            else:
                try:
                    scan, priorities, product_scans = self._process_ms1(
                        scan, product_scans)
                    if scan is None:
                        # no way to report skip
                        pass
                    else:
                        if self.verbose:
                            self.log_message("Handling Precursor Scan %r with %d peaks" % (scan.id, len(scan.peak_set)))
                        if self.deconvolute:
                            self._deconvolute_ms1(scan, priorities, product_scans)
                        self.send_scan(scan)
                except (KeyboardInterrupt, SystemExit) as e:
                    raise
                except NoIsotopicClustersError as e:
                    self.log_message("No isotopic clusters were extracted from scan %s (%r peaks)" % (
                        e.scan_id, len(scan.peak_set)))
                    self.skip_scan(scan)
                except EmptyScanError as e:
                    self.skip_scan(scan)
                except Exception as e:
                    self.skip_scan(scan)
                    self.log_error(e, scan_id, scan, (product_scan_ids))
        except (KeyboardInterrupt, SystemExit) as e:
            raise
        except Exception as err:
            self.skip_scan(scan)
            self.log_error(err, scan_id, scan, product_scan_ids)
        return scan, product_scans

    def _process_msn(self, product_scan: Scan):
        self.transformer.pick_product_scan_peaks(product_scan)

    def _deconvolute_msn(self, product_scan: Scan):
        self.transformer.deconvolute_product_scan(product_scan)

    def _handle_msn(self, product_scan: Scan, precursor_scan: Scan):
        try:
            self._process_msn(product_scan)
            if self.verbose:
                self.log_message("Handling Product Scan %r with %d peaks (%0.3f/%0.3f, %r)" % (
                    product_scan.id, len(product_scan.peak_set), product_scan.precursor_information.mz,
                    product_scan.precursor_information.extracted_mz,
                    product_scan.precursor_information.defaulted))
            if self.deconvolute:
                self._deconvolute_msn(product_scan)
                if precursor_scan is None and product_scan.precursor_information:
                    product_scan.precursor_information.default(orphan=True)
            self.send_scan(product_scan)
        except (KeyboardInterrupt, SystemExit) as e:
            raise
        except NoIsotopicClustersError as e:
            self.log_message("No isotopic clusters were extracted from scan %s (%r)" % (
                e.scan_id, len(product_scan.peak_set)))
            self.skip_scan(product_scan)
        except EmptyScanError as e:
            self.skip_scan(product_scan)
        except Exception as e:
            self.skip_scan(product_scan)
            self.log_error(e, product_scan.id, product_scan, ())

    def handle_scan_bunch(self, scan: Scan, product_scans: List[Scan], scan_id: str, product_scan_ids: List[str], process_msn: bool=True):
        transformer = self.transformer
        # handle the MS1 scan if it is present
        if scan is not None:
            scan, product_scans = self._handle_ms1_scan(
                scan, product_scans, scan_id, product_scan_ids)
        for product_scan in product_scans:
            # no way to report skip
            try:
                if product_scan is None:
                    continue
                if len(product_scan.arrays[0]) == 0 or (not process_msn):
                    self.skip_scan(product_scan)
                    continue
                self._handle_msn(product_scan, scan)
            except (KeyboardInterrupt, SystemExit) as e:
                raise
            except Exception as err:
                self.skip_scan(product_scan)
                self.log_error(err, product_scan.id,
                               product_scan, [])

    def _open_ms_file(self) -> Union[RandomAccessScanSource, ScanIterator]:
        self.loader = MSFileLoader(self.ms_file_path, decode_binary=False)
        return self.loader

    def _make_batch_loader(self, loader: Union[ScanIterator, RandomAccessScanSource]) -> ScanBunchLoader:
        return ScanBunchLoader(loader)

    def run(self):
        self.try_set_process_name("ms-deisotope-deconv")
        self._silence_loggers()
        loader = self._open_ms_file()
        queued_loader = self._make_batch_loader(loader)
        transformer = self.make_scan_transformer(loader)

        has_input: bool = True
        i: int = 0
        last: int = 0
        scan = None
        product_scans = []
        while has_input:
            try:
                scan_id, product_scan_ids, process_msn = self.get_work(True, 10)
            except QueueEmpty:
                if self.no_more_event is not None and self.no_more_event.is_set():
                    has_input = False
                continue

            i += 1 + len(product_scan_ids)
            if scan_id == DONE:
                has_input = False
                break

            try:
                queued_loader.put(scan_id, product_scan_ids)
                scan, product_scans = queued_loader.get()
            except (KeyboardInterrupt, SystemExit) as e:
                self.log_message("Interrupt received.")
                break
            except Exception as e:
                self.log_message("Something went wrong when loading bunch (%s): %r.\nRecovery is not possible." % (
                    (scan_id, product_scan_ids), e))

            self.handle_scan_bunch(
                scan, product_scans,
                scan_id, product_scan_ids,
                process_msn)

            if (i - last) > 1000:
                last = i
                self.output_queue.join()

        self.log_message("Done (%d scans)" % i)

        if self.no_more_event is None:
            self.output_queue.put((DONE, DONE, DONE))

        if os.getenv("MS_DEISOTOPE_DEBUG"):
            self._dump_averagine_caches(transformer)

        self._work_complete.set()

    def _dump_averagine_caches(self, transformer: ScanProcessor):
        pid = os.getpid()
        fname = "ms_deisotope_averagine_cache_ms$_%s.pkl" % pid
        with open(fname.replace("$", '1'), 'wb') as fh:
            pickle.dump(transformer.ms1_deconvolution_args.get('averagine'), fh, 2)
        with open(fname.replace("$", 'n'), 'wb') as fh:
            pickle.dump(transformer.msn_deconvolution_args.get('averagine'), fh, 2)




from collections import deque
from typing import Any, Deque, Iterator, List, Optional, Set, Type, Sequence

import numpy as np

from ms_deisotope.peak_dependency_network.intervals import (
    SpanningMixin,
    IntervalTreeNode,
    Interval,
    SimpleInterval,
    IntervalTreeNode2D
)

from ms_deisotope.feature_map.lcms_feature import LCMSFeature
from ._mz_feature_search import MZIndex, NeutralMassIndex


def mz_point_organizer_callback(contained_intervals):
    return IntervalTreeNode.build(
        [Interval(node.mz, node.mz, [node]) for node in contained_intervals])


def neutral_mass_point_organizer_callback(contained_intervals):
    return IntervalTreeNode.build(
        [Interval(node.neutral_mass, node.neutral_mass, [node]) for node in contained_intervals])


class FeatureGraphNode(SpanningMixin):
    feature: LCMSFeature
    index: int
    edges: Set['FeatureGraphEdge']
    center: float
    mz: float

    def __init__(self, feature: LCMSFeature, index: int, edges: Optional[Set]=None):
        self.start = feature.start_time
        self.end = feature.end_time
        if edges is None:
            edges = set()
        self.feature = feature
        self.index = index
        self.edges = edges

        total = 0
        abundance = 0
        for node in self.feature.nodes:
            intensity = node.total_intensity()
            total += node.time * intensity
            abundance += intensity
        self.center = total / abundance
        self.mz = feature.mz

    @property
    def start_time(self) -> float:
        return self.start

    @property
    def end_time(self) -> float:
        return self.end

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.feature,)

    def __index__(self):
        return self.index

    def __hash__(self):
        return hash(self.index)

    def __eq__(self, other):
        return self.feature == other.feature


class DeconvolutedFeatureGraphNode(FeatureGraphNode):
    neutral_mass: float
    charge: int

    def __init__(self, feature, index, edges=None):
        super().__init__(feature, index, edges)
        self.neutral_mass = feature.neutral_mass
        self.charge = feature.charge


class IonMobilityDeconvolutedFeatureGraphNode(DeconvolutedFeatureGraphNode):
    ion_mobility: float

    def __init__(self, feature, index, edges=None):
        super().__init__(feature, index, edges)
        self.ion_mobility = feature.drift_time


class IonMobilityProfileFeatureGraphNode(DeconvolutedFeatureGraphNode):
    ion_mobility_interval: SimpleInterval

    def __init__(self, feature, index, edges=None):
        super().__init__(feature, index, edges)
        self.ion_mobility_interval = feature.ion_mobility_interval


class FeatureGraphEdge(object):
    __slots__ = ('node_a', 'node_b', 'transition', 'weight', 'mass_error', 'rt_error', '_hash')

    node_a: FeatureGraphNode
    node_b: FeatureGraphNode
    transition: Any
    weight: float
    mass_error: float
    rt_error: float
    _hash: int

    def __init__(self, node_a, node_b, transition, weight=1, mass_error=0, rt_error=0):
        self.node_a = node_a
        self.node_b = node_b
        self.transition = transition
        self.weight = weight
        self.mass_error = mass_error
        self.rt_error = rt_error
        if self.node_a.index < self.node_b.index:
            self._hash = hash(((self.node_a.index, self.node_b.index)))
        else:
            self._hash = hash(((self.node_b.index, self.node_a.index)))
        self.node_a.edges.add(self)
        self.node_b.edges.add(self)

    def __repr__(self):
        return "%s(%s, %s, %s)" % (
            self.__class__.__name__, self.node_a.feature, self.node_b.feature,
            self.transition)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return self.node_a.index == other.node_a.index and self.node_b.index == other.node_b.index

    def __ne__(self, other):
        return not self == other

    def remove(self):
        self.node_a.edges.remove(self)
        self.node_b.edges.remove(self)


class TimeQuery(SpanningMixin):
    __slots__ = ()

    def __init__(self, feature: LCMSFeature, width: float = 0):
        self.start = feature.start_time - width
        self.end = feature.end_time + width

    def __repr__(self):
        return "TimeQuery(%f, %f)" % (self.start, self.end)


class PPMQuery(SpanningMixin):
    __slots__ = ()

    def __init__(self, center, error_tolerance):
        self.start = center - center * error_tolerance
        self.end = center + center * error_tolerance

    def __repr__(self):
        return "PPMQuery(%f, %f)" % (self.start, self.end)


try:
    from ms_deisotope._c.feature_map.feature_graph import (
        FeatureGraphNode, DeconvolutedFeatureGraphNode, IonMobilityProfileFeatureGraphNode,
        FeatureGraphEdge)
except ImportError:
    raise


class FeatureGraph(MZIndex):
    node_cls: Type = FeatureGraphNode
    edge_cls: Type = FeatureGraphEdge

    _features: Sequence[LCMSFeature]
    nodes: List[FeatureGraphNode]
    assigned_seed_queue: Deque[FeatureGraphNode]
    rt_tree: IntervalTreeNode2D
    edges: Set[FeatureGraphEdge]

    @property
    def features(self):
        return self.nodes

    def __init__(self, features):
        self._features = features
        self.assigned_seed_queue = deque()
        self.nodes = self._construct_graph_nodes(self._features)
        self.edges = set()
        self.rt_tree = None
        self._make_rt_tree()

    def _make_rt_tree(self):
        self.rt_tree = IntervalTreeNode2D.build(
            self.nodes, mz_point_organizer_callback)

    def __len__(self):
        return len(self.nodes)

    def __iter__(self):
        return iter(self.nodes)

    def __getitem__(self, i):
        return self.nodes[i]

    def _construct_graph_nodes(self, features):
        nodes = []
        for i, chroma in enumerate(features):
            node = (self.node_cls(chroma, i))
            nodes.append(node)
            self.enqueue_seed(node)
        return nodes

    def enqueue_seed(self, feature):
        self.assigned_seed_queue.append(feature)

    def pop_seed(self) -> FeatureGraphNode:
        return self.assigned_seed_queue.popleft()

    def iterseeds(self) -> Iterator[FeatureGraphNode]:
        while self.assigned_seed_queue:
            yield self.pop_seed()

    def find_edges(self, node: FeatureGraphNode, query_width: float = 2., error_tolerance=1.5e-5, **kwargs):
        query = TimeQuery(node.feature, query_width)
        mass_query = PPMQuery(node.mz, error_tolerance)
        nodes = self.rt_tree.overlaps_2d(
            np.array([query.start, mass_query.start]), np.array([query.end, mass_query.end]))

        for match in nodes:
            ppm_error = (node.mz - match.mz) / match.mz
            if abs(ppm_error) > error_tolerance:
                continue
            if match.index == node.index:
                continue
            rt_error = (node.center - match.center)
            self.edges.add(self.edge_cls(
                node, match, None, mass_error=ppm_error, rt_error=rt_error))

    def build(self, query_width: float = 2., error_tolerance=1.5e-5, **kwargs):
        for node in self.iterseeds():
            self.find_edges(node, query_width=query_width,
                            error_tolerance=error_tolerance, **kwargs)

    def adjacency_matrix(self):
        adjmat = np.zeros((len(self), len(self)))
        nodes = self.nodes
        for node in nodes:
            for edge in node.edges:
                adjmat[edge.node_a.index, edge.node_b.index] = 1
        return adjmat

    def connected_components(self) -> List[List[FeatureGraphNode]]:
        pool = set(self.nodes)
        components = []

        i = 0
        while pool:
            i += 1
            current_component = set()
            visited = set()
            node = pool.pop()
            current_component.add(node)
            j = 0
            while current_component:
                j += 1
                node = current_component.pop()
                visited.add(node)
                for edge in node.edges:
                    if edge.node_a not in visited and edge.node_a in pool:
                        current_component.add(edge.node_a)
                        pool.remove(edge.node_a)
                    if edge.node_b not in visited and edge.node_b in pool:
                        current_component.add(edge.node_b)
                        pool.remove(edge.node_b)
            components.append(list(visited))
        return components


class DeconvolutedFeatureGraph(FeatureGraph, NeutralMassIndex):
    node_cls = DeconvolutedFeatureGraphNode

    def _make_rt_tree(self):
        self.rt_tree = IntervalTreeNode2D.build(
            self.nodes, neutral_mass_point_organizer_callback)

    def find_all(self, mass: float, error_tolerance: float=0.00002):
        return NeutralMassIndex.find_all(self, mass, error_tolerance)

    def find_edges(self, node: DeconvolutedFeatureGraphNode, query_width: float = 2., error_tolerance: float=1.5e-5, **kwargs):
        query = TimeQuery(node.feature, query_width)
        mass_query = PPMQuery(node.neutral_mass, error_tolerance)
        nodes = self.rt_tree.overlaps_2d(
            np.array([query.start, mass_query.start]), np.array([query.end, mass_query.end]))
        charge = node.charge
        for match in nodes:
            if match.charge != charge:
                continue
            ppm_error = (node.neutral_mass - match.neutral_mass) / match.neutral_mass
            if abs(ppm_error) > error_tolerance:
                continue
            if match.index == node.index:
                continue
            rt_error = (node.center - match.center)
            self.edges.add(self.edge_cls(
                node, match, None, mass_error=ppm_error, rt_error=rt_error))


class IonMobilityDeconvolutedFeatureGraph(DeconvolutedFeatureGraph):
    node_cls = IonMobilityDeconvolutedFeatureGraphNode

    def find_edges(self, node: IonMobilityDeconvolutedFeatureGraphNode, query_width: float = 2., error_tolerance: float = 1.5e-5, ion_mobility_error_tolerance: float = 0.01, **kwargs):
        query = TimeQuery(node.feature, query_width)
        mass_query = PPMQuery(node.neutral_mass, error_tolerance)
        nodes = self.rt_tree.overlaps_2d(
            np.array([query.start, mass_query.start]), np.array([query.end, mass_query.end]))
        charge = node.charge
        for match in nodes:
            if match.charge != charge:
                continue
            # If there isn't an overlap in the IM dimension, then they aren't the same molecule even
            # if charge, mass, and elution time do match
            if abs(match.ion_mobility - node.ion_mobility) > ion_mobility_error_tolerance:
                continue
            ppm_error = (node.neutral_mass - match.neutral_mass) / \
                match.neutral_mass
            if abs(ppm_error) > error_tolerance:
                continue
            if match.index == node.index:
                continue
            rt_error = (node.center - match.center)
            self.edges.add(self.edge_cls(
                node, match, None, mass_error=ppm_error, rt_error=rt_error))


class IonMobilityProfileDeconvolutedFeatureGraph(DeconvolutedFeatureGraph):
    node_cls = IonMobilityProfileFeatureGraphNode

    def find_edges(self, node: IonMobilityProfileFeatureGraphNode, query_width: float = 2., error_tolerance: float=1.5e-5, **kwargs):
        query = TimeQuery(node.feature, query_width)
        mass_query = PPMQuery(node.neutral_mass, error_tolerance)
        nodes = self.rt_tree.overlaps_2d(
            np.array([query.start, mass_query.start]), np.array([query.end, mass_query.end]))
        charge = node.charge
        for match in nodes:
            if match.charge != charge:
                continue
            # If there isn't an overlap in the IM dimension, then they aren't the same molecule even
            # if charge, mass, and elution time do match
            if not match.ion_mobility_interval.overlaps(node.ion_mobility_interval):
                continue
            ppm_error = (node.neutral_mass - match.neutral_mass) / \
                match.neutral_mass
            if abs(ppm_error) > error_tolerance:
                continue
            if match.index == node.index:
                continue
            rt_error = (node.center - match.center)
            self.edges.add(self.edge_cls(
                node, match, None, mass_error=ppm_error, rt_error=rt_error))


try:
    from ms_deisotope._c.feature_map.feature_graph import (
        find_edges, find_edges_neutral_mass, find_edges_neutral_mass_ion_mobility, connected_components)

    FeatureGraph.find_edges = find_edges
    FeatureGraph.connected_components = connected_components
    DeconvolutedFeatureGraph.find_edges = find_edges_neutral_mass
    IonMobilityProfileDeconvolutedFeatureGraph.find_edges = find_edges_neutral_mass_ion_mobility
except ImportError:
    raise


class GapAwareFeatureSmoother(object):
    graph: FeatureGraph

    def __init__(self, features):
        self.graph = FeatureGraph(features)

    def _wrap_result(self, features) -> 'LCMSFeatureMap':
        from ms_deisotope.feature_map.feature_map import LCMSFeatureMap
        return LCMSFeatureMap(features)

    def connect_components(self, time_bridge: float=0.5, mass_error_tolerance: float=1.5e-5):
        self.graph.build(query_width=time_bridge, error_tolerance=mass_error_tolerance)
        features: List[LCMSFeature] = []
        for component in self.graph.connected_components():
            f = component[0].feature
            for f2 in component[1:]:
                f = f.merge(f2.feature)
            features.append(f)
        return self._wrap_result(features)

    @classmethod
    def smooth(cls, features, time_bridge: float=0.5, mass_error_tolerance: float=1.5e-5, **kwargs):
        return cls(features).connect_components(time_bridge=time_bridge, mass_error_tolerance=mass_error_tolerance, **kwargs)


class GapAwareDeconvolutedFeatureSmoother(GapAwareFeatureSmoother):
    graph: DeconvolutedFeatureGraph

    def __init__(self, features):
        self.graph = DeconvolutedFeatureGraph(features)

    def _wrap_result(self, features) -> 'DeconvolutedLCMSFeatureMap':
        from ms_deisotope.feature_map.feature_map import DeconvolutedLCMSFeatureMap
        return DeconvolutedLCMSFeatureMap(features)


class GapAwareIonMobilityDeconvolutedFeatureSmoother(GapAwareDeconvolutedFeatureSmoother):
    graph: IonMobilityDeconvolutedFeatureGraph

    def __init__(self, features):
        self.graph = IonMobilityDeconvolutedFeatureGraph(features)

    def connect_components(self, time_bridge: float = 0.5, mass_error_tolerance: float = 1.5e-5, ion_mobility_error_tolerance: float=0.1):
        self.graph.build(query_width=time_bridge,
                         error_tolerance=mass_error_tolerance,
                         ion_mobility_error_tolerance=ion_mobility_error_tolerance)
        features: List[LCMSFeature] = []
        for component in self.graph.connected_components():
            f = component[0].feature
            for f2 in component[1:]:
                f = f.merge(f2.feature)
            features.append(f)
        return self._wrap_result(features)

    def _wrap_result(self, features) -> 'IonMobilityDeconvolutedLCMSFeatureMap':
        from ms_deisotope.feature_map.feature_map import IonMobilityDeconvolutedLCMSFeatureForest
        return IonMobilityDeconvolutedLCMSFeatureForest(features)


class GapAwareIonMobilityProfileDeconvolutedFeatureSmoother(GapAwareDeconvolutedFeatureSmoother):
    graph: IonMobilityProfileDeconvolutedFeatureGraph

    def __init__(self, features):
        self.graph = IonMobilityProfileDeconvolutedFeatureGraph(features)



import io

DEFAULT_BUFFER_SIZE = 2 ** 18


class PreBufferedStreamReader(io.IOBase):
    """A file-like object that can wrap a non-rewindable (e.g. un-seekable)
    file stream like :obj:`sys.stdin` and support seeking over the first **N**
    bytes, but blocking seek operations once the stream has been read beyond
    that point.

    This type exists to adapt :obj:`sys.stdin` or socket files to work with the
    behavior of readers like:class:`~ms_deisotope.data_source.mzml.MzMLLoader`
    that rely on early random access to extract metadata even when full random
    access indexing isn't enabled.

    Attributes
    ----------
    stream : file-like object
        The object to actually read from. Must support :meth:`io.IOBase.read`,
        :meth:`io.IOBase
    """
    def __init__(self, stream, buffer_size=DEFAULT_BUFFER_SIZE):
        self.stream = stream
        dat = self.stream.read(buffer_size)
        self.buffer = io.BytesIO(dat)
        self._cursor = 0
        self._total_cursor = len(dat)
        self._buffer_size = buffer_size

    def read(self, n=-1):
        if n == -1:
            buffer = bytearray()
            buffer.extend(self.buffer.read())
            buffer.extend(self.stream.read())
            self._cursor = self._buffer_size
            self._total_cursor = len(buffer) - self._cursor
            return bytes(buffer)
        buffer = bytearray(n)
        view = memoryview(buffer)
        k = self.buffer.readinto(view)
        self._cursor += k
        j = 0
        if k < n:
            j = self.stream.readinto(view[k:])
            self._total_cursor += j
            k += j
        return bytes(view[:k])

    def fileno(self):
        return self.stream.fileno()

    def seek(self, offset, whence=io.SEEK_SET):
        if (whence == io.SEEK_SET and offset > self._buffer_size):
            raise io.UnsupportedOperation(
                "Cannot seek beyond the pre-loaded buffer")
        elif whence == io.SEEK_CUR or whence == io.SEEK_END:
            raise io.UnsupportedOperation(
                "Cannot seek not from the start of the file")
        if self._total_cursor > self._buffer_size:
            raise io.UnsupportedOperation(
                "Already read beyond the pre-loaded buffer")
        else:
            self.buffer.seek(offset, whence)
            self._cursor = offset

    def truncate(self, size=None):
        raise io.UnsupportedOperation("Read-only")

    def close(self):
        self.buffer.close()
        return self.stream.close()

    def closed(self):
        return self.stream.closed()

    def seekable(self):
        return True

    def tell(self):
        if self._cursor < self._buffer_size:
            return self._cursor
        return self._total_cursor

    def writable(self):
        return False

    def __enter__(self):
        self.stream.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


"""
The Windows DLL needed for the MassLynx reader to work can be obtained from:
https://interface.waters.com/masslynx/developers-area/sdks/, and by linking
with them we agree to follow the associated licensing.
"""

import os
from ctypes import WinDLL
from ctypes.util import find_library

dll = None


def register_dll(search_paths=None, override=True):
    result = _register_dll(search_paths, override)
    if not result:
        raise ImportError("Could not load MassLynx SDK from search paths!")



def _register_dll(search_paths=None, override=True):
    if search_paths is None:
        search_paths = []
    elif isinstance(search_paths, str):
        search_paths = [search_paths]
    elif not isinstance(search_paths, list):
        raise TypeError("Expected a list of paths or a single path string")
    from ms_deisotope.config import get_config
    search_paths = list(search_paths)
    search_paths.extend(get_config().get('vendor_readers', {}).get('waters-masslynx', []))
    found = find_library("MassLynx.dll")
    if found:
        search_paths.append(found)
    global dll
    if dll is None or override:
        for lib_path in search_paths:
            try:
                dll = _load_library(lib_path)
            except (Exception) as err:
                continue
            if dll is not None:
                break
    return dll


def _load_library(lib_path):
    dll = WinDLL(os.path.realpath(lib_path))
    return dll


def determine_if_available():
    """Checks whether or not the Waters
    RAW directory reading feature is available.

    Returns
    -------
    :class:`bool`:
        Whether or not the feature is enabled.
    """
    try:
        return _register_dll(override=False)
    except (OSError, ImportError):
        return False


class DLLProxy(object):
    def __getattr__(self, attribute):
        if dll is None:
            register_dll()
        return getattr(dll, attribute)

    def _is_loaded(self):
        return dll is not None

proxy = DLLProxy()



cimport cython

from ms_deisotope._c.peak_set cimport DeconvolutedPeak


cdef class NodeBase(object):

    cdef:
        public size_t key

    cdef double get_neutral_mass(self)
    cdef double get_intensity(self)
    cdef size_t get_index(self)
    cdef bint _eq(self, NodeBase other)


cdef class PlaceHolderNode(NodeBase):
    cdef:
        public str label
        public double _mass
        public double _intensity
        public size_t _index


cdef class PeakNode(NodeBase):
    cdef:
        public DeconvolutedPeak peak



cdef class PeakGroupNode(NodeBase):
    cdef:
        public list peaks

    cdef size_t get_size(self)


@cython.final
@cython.freelist(1000000)
cdef class Edge(object):
    cdef:
        public NodeBase start
        public NodeBase end
        public object annotation
        public (size_t, size_t) key
        public Py_hash_t _hash

    @staticmethod
    cdef Edge _create(NodeBase start, NodeBase end, object annotation)


cdef class SpectrumGraph(object):
    cdef:
        public set transitions
        public object by_first
        public object by_second

    cpdef add(self, DeconvolutedPeak p1, DeconvolutedPeak p2, object annotation)
    cpdef list topological_sort(self, adjacency_matrix=*)

    cpdef list paths_starting_at(self, ix, int limit=*)
    cpdef list paths_ending_at(self, ix, int limit=*)

    cpdef list transitive_closure(self, list paths, int limit=*)


cdef class MassWrapper(object):
    """An adapter class to make types whose mass calculation is a method
    (:mod:`glypy` dynamic graph components) compatible with code where the
    mass calculation is an  attribute (:mod:`glycopeptidepy` objects and
    most things here)

    Hashes and compares as :attr:`obj`

    Attributes
    ----------
    obj: object
        The wrapped object
    mass: float
        The mass of :attr:`obj`
    """
    cdef:
        public object obj
        public double mass


cdef class Path(object):
    cdef:
        public list transitions
        public double total_signal
        public double start_mass
        public double end_mass
        public set _peaks_used
        public set _edges_used

    cdef double _total_signal(self)
    cdef Edge get(self, Py_ssize_t i)
    cdef size_t get_size(self)


"""
     Waters
    MassLynx Python SDK
"""

import ctypes
from ctypes import*

from .MassLynxRawReader import MassLynxRawReader
from .MassLynxRawDefs import MassLynxBaseType
from .MassLynxParameters import MassLynxParameters

from enum import IntEnum


class MassLynxRawInfoReader(MassLynxRawReader):

    def __init__(self, source ):
        super(MassLynxRawInfoReader, self).__init__(source, MassLynxBaseType.INFO)


    def GetNumberofFunctions( self ):
        size = c_int(0)
        getFunctionCount =  MassLynxRawReader.massLynxDll.getFunctionCount
        getFunctionCount.argtypes = [c_void_p, POINTER(c_int)]
        self.CheckReturnCode( getFunctionCount(self._getReader(),size) )

        return size.value

    def GetScansInFunction( self, whichFunction ):
        size = c_int(0)
        getScanCount = MassLynxRawReader.massLynxDll.getScanCount
        getScanCount.argtypes = [c_void_p, c_int, POINTER(c_int)]
        self.CheckReturnCode( getScanCount(self._getReader(),whichFunction,size) )

        return size.value

    def GetAcquisitionMassRange( self, whichFunction ):
        lowMass = c_float(0)
        highMass = c_float(0)
        getAcquisitionMassRange = MassLynxRawReader.massLynxDll.getAcquisitionMassRange
        getAcquisitionMassRange.argtypes = [c_void_p, c_int, c_int, POINTER(c_float), POINTER(c_float)]
        self.CheckReturnCode( getAcquisitionMassRange(self._getReader(),whichFunction, 0,lowMass,highMass) )

        return lowMass.value, highMass.value

    def GetAcquisitionTimeRange( self, whichFunction ):
        startTime = c_float(0)
        endTime = c_float(0)
        getAcquisitionTimeRange = MassLynxRawReader.massLynxDll.getAcquisitionTimeRange
        getAcquisitionTimeRange.argtypes = [c_void_p, c_int, POINTER(c_float), POINTER(c_float)]
        self.CheckReturnCode( getAcquisitionTimeRange(self._getReader(),whichFunction,startTime,endTime) )

        return startTime.value, endTime.value

    def GetFunctionType( self, whichFunction ):
        functionType = c_int(0)
        getFunctionType = MassLynxRawReader.massLynxDll.getFunctionType
        getFunctionType.argtypes = [c_void_p, c_int, POINTER(c_int)]
        self.CheckReturnCode( getFunctionType(self._getReader(),whichFunction, functionType) )

        return functionType.value

    def GetFunctionTypeString( self, functionType ):
        functionTypeString = c_char_p()
        temp = c_char_p()
        getFunctionTypeString = MassLynxRawReader.massLynxDll.getFunctionTypeString
        getFunctionTypeString.argtypes = [c_void_p, c_int, POINTER(c_char_p)]
        self.CheckReturnCode( getFunctionTypeString(self._getReader(),functionType, temp) )

        return self.ToString( temp )

    def IsContinuum( self, whichFunction ):
        continuum = c_bool(0)
        isContinuum = MassLynxRawReader.massLynxDll.isContinuum
        isContinuum.argtypes = [c_void_p, c_int, POINTER(c_bool)]
        self.CheckReturnCode( isContinuum(self._getReader(),whichFunction, continuum) )

        return continuum.value

    def GetIonMode( self, whichFunction ):
        ionMode = c_int()
        getIonMode = MassLynxRawReader.massLynxDll.getIonMode
        getIonMode.argtypes = [c_void_p, c_int, POINTER(c_int)]
        self.CheckReturnCode(getIonMode(self._getReader(),whichFunction, ionMode ))

        return ionMode.value

    def GetIonModeString( self, ionMode ):
        ionModeString = c_char_p()
        temp = c_char_p()
        getIonModeString = MassLynxRawReader.massLynxDll.getIonModeString
        getIonModeString.argtypes = [c_void_p, c_int, POINTER(c_char_p)]
        self.CheckReturnCode(getIonModeString(self._getReader(),ionMode,  temp ))
        return self.ToString(temp)

    def GetHeaderItems( self, whichItems ):
        itemString = c_char_p()
        nItems = len(whichItems )

        temp = c_char_p()
        items = (c_int * nItems)(*whichItems)
        delimiter = ctypes.create_string_buffer(1)

        getHeaderItem = MassLynxRawReader.massLynxDll.getHeaderItems
        getHeaderItem.argtypes = [c_void_p, POINTER(c_int), POINTER(c_char_p), c_int, POINTER(c_char)]
        self.CheckReturnCode( getHeaderItem( self._getReader(), items, temp, nItems, delimiter))

        itemString = self.ToString(temp)

        delim = delimiter.value.decode()
        return itemString.split(delim);

    def GetHeaderItemValue( self, whichItems ):
        nItems = len(whichItems )
        items = (c_int * nItems)(*whichItems)
        params = MassLynxParameters()
        getHeaderItemValue = MassLynxRawReader.massLynxDll.getHeaderItemValue
        getHeaderItemValue.argtypes = [c_void_p, POINTER(c_int), c_int, c_void_p]
        self.CheckReturnCode(getHeaderItemValue(self._getReader(), items, nItems, params.GetParameters()))

        return params

    def GetHeaderItem( self, whichItem ):
        whichItems =  list()
        whichItems.append( whichItem )
        values =  self.GetHeaderItems( whichItems)

        return values[0]

    # scan stats
    def GetScanItem( self, whichFunction, whichScan, whichItem ):
        whichItems = list()
        whichItems.append( whichItem )
        values =  self.GetScanItems(whichFunction, whichScan, whichItems)

        return values[0]

    def GetScanItems(self, whichFunction, whichScan, whichItems):
        nItems = len(whichItems)
        temp = c_char_p()
        items = (c_int * nItems)(*whichItems)
        delimiter = ctypes.create_string_buffer(1)

        getScanItem = MassLynxRawReader.massLynxDll.getScanItems
        getScanItem.argtypes = [c_void_p, c_int, c_int, POINTER(c_int), POINTER(c_char_p), c_int, POINTER(c_char)]
        self.CheckReturnCode( getScanItem( self._getReader(), whichFunction, whichScan, items, temp, nItems, delimiter))

        itemString = self.ToString(temp)

        delim = delimiter.value.decode()
        return itemString.split(delim);

    def GetScanItemString(self, whichItems):
         # get the array of items
        nItems = len(whichItems)
        temp = c_char_p()
        items = (c_int * nItems)(*whichItems)
        delimiter = ctypes.create_string_buffer(1)

        getScanItemNames = MassLynxRawReader.massLynxDll.getScanItemNames
        getScanItemNames.argtypes = [c_void_p, POINTER(c_int), POINTER(c_char_p), c_int, POINTER(c_char)]
        self.CheckReturnCode( getScanItemNames( self._getReader(), items, temp, nItems, delimiter))

        itemString = self.ToString(temp)

        delim = delimiter.value.decode()
        return itemString.split(delim);


    def GetItemsInFunction( self, whichFunction, nWhichScan ):
        size = c_int(0)
        pItems = c_void_p()
        getItemsInFunction = MassLynxRawReader.massLynxDll.getItemsInFunction
        getItemsInFunction.argtypes = [c_void_p, c_int, POINTER(c_void_p), POINTER(c_int)]
        self.CheckReturnCode( getItemsInFunction(self._getReader(),whichFunction,pItems,size) )

        # fill the array
        pI = cast(pItems,POINTER(c_int))
        items = pI[0:size.value]

        # dealocate memory
        MassLynxRawReader.ReleaseMemory( pItems)

        return items

    def GetRetentionTime( self, whichFunction, nWhichScan ):
        retentionTime = c_float(0)
        getRetentionTime = MassLynxRawReader.massLynxDll.getRetentionTime
        getRetentionTime.argtypes = [c_void_p, c_int, c_int, POINTER(c_float)]
        self.CheckReturnCode( getRetentionTime(self._getReader(),whichFunction,nWhichScan,retentionTime) )

        return retentionTime.value

    def GetDriftTime( self, whichFunction, nWhichDrift ):
        driftTime = c_float(0)
        getDriftTime = MassLynxRawReader.massLynxDll.getDriftTime
        getDriftTime.argtypes = [c_void_p, c_int, c_int, POINTER(c_float)]
        self.CheckReturnCode( getDriftTime(self._getReader(),whichFunction,nWhichDrift,driftTime) )

        return driftTime.value

    def GetDriftScanCount(self, whichFunction):
        scan_count = c_int(0)
        getDriftScanCount = MassLynxRawReader.massLynxDll.getDriftScanCount
        getDriftScanCount.argtypes = [c_void_p, c_int, POINTER(c_int)]
        self.CheckReturnCode(getDriftScanCount(self._getReader(), whichFunction, scan_count))
        return scan_count.value

    def GetCollisionalCrossSection(self, drift_time, mass, charge):
        ccs = c_float(0)
        getCollisionalCrossSection = MassLynxRawReader.massLynxDll.getCollisionalCrossSection
        getCollisionalCrossSection.argtypes = [
            c_void_p, c_float, c_float, c_int, POINTER(c_float)]
        self.CheckReturnCode(getCollisionalCrossSection(self._getReader(), drift_time, mass, charge, ccs))
        return ccs.value

    def CanLockMassCorrect( self ):
        canApply = c_bool(0)
        canLockMassCorrect =  MassLynxRawReader.massLynxDll.canLockMassCorrect
        canLockMassCorrect.argtypes = [c_void_p, POINTER(c_bool)]
        self.CheckReturnCode( canLockMassCorrect(self._getReader(), canApply) )

        return canApply.value


    def IsLockMassCorrected( self ):
        corrected = c_bool(0)
        mlMethod =  MassLynxRawReader.massLynxDll.isLockMassCorrected
        mlMethod.argtypes = [c_void_p, POINTER(c_bool)]
        self.CheckReturnCode( mlMethod(self._getReader(), corrected) )

        return corrected.value


